"""
Tools for modeling "rhythm-trees".
"""

import abc
import typing
import uqbar.containers
import uqbar.graphs
from abjad import Fraction
from abjad import core
from abjad import mathtools
from abjad import system
from abjad.system import Parser


class RhythmTreeMixin(system.AbjadObject):
    """
    Abstract rhythm-tree node.
    """

    ### CLASS VARIABLES ###

    _state_flag_names: typing.Tuple[str, ...] = ('_offsets_are_current',)

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, preprolated_duration=1):
        import abjad
        self._duration = 0
        self._offset = abjad.Offset(0)
        self._offsets_are_current = False
        self.preprolated_duration = preprolated_duration

    ### SPECIAL METHODS ###

    @abc.abstractmethod
    def __call__(self, pulse_duration):
        """
        Calls rhythm tree node on ``pulse_duration``.
        """
        raise NotImplementedError

    ### PRIVATE METHODS ###

    def _update_offsets_of_entire_tree(self):
        def recurse(container, current_offset):
            container._offset = current_offset
            container._offsets_are_current = True
            for child in container:
                if getattr(child, 'children', None) is not None:
                    current_offset = recurse(child, current_offset)
                else:
                    child._offset = current_offset
                    child._offsets_are_current = True
                    current_offset += child.duration
            return current_offset
        import abjad
        offset = abjad.Offset(0)
        root = self.root
        if root is None:
            root = self
        if root is self and not hasattr(self, 'children'):
            self._offset = offset
            self._offsets_are_current = True
        else:
            recurse(root, offset)

    def _update_offsets_of_entire_tree_if_necessary(self):
        if not self._get_node_state_flags()['_offsets_are_current']:
            self._update_offsets_of_entire_tree()

    ### PRIVATE PROPERTIES ###

    @abc.abstractproperty
    def _pretty_rtm_format_pieces(self):
        raise NotImplementedError

    @property
    def _depthwise_inventory(self):
        def recurse(node):
            if node.depth not in inventory:
                inventory[node.depth] = []
            inventory[node.depth].append(node)
            if getattr(node, 'children', None) is not None:
                for child in node.children:
                    recurse(child)
        inventory = {}
        recurse(self)
        return inventory

    ### PUBLIC PROPERTIES ###

    @property
    def duration(self):
        """
        The preprolated_duration of the node:

        >>> rtm = '(1 ((1 (1 1)) (1 (1 1))))'
        >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]

        >>> tree.duration
        Duration(1, 1)

        >>> tree[1].duration
        Duration(1, 2)

        >>> tree[1][1].duration
        Duration(1, 4)

        Return ``Duration`` instance.
        """
        return self.prolation * self.preprolated_duration

    @property
    def parentage_ratios(self):
        """
        A sequence describing the relative durations of the nodes in a
        node's improper parentage.

        The first item in the sequence is the preprolated_duration of
        the root node, and subsequent items are pairs of the
        preprolated duration of the next node in the parentage and
        the total preprolated_duration of that node and its siblings:


        >>> a = abjad.rhythmtrees.RhythmTreeContainer(preprolated_duration=1)
        >>> b = abjad.rhythmtrees.RhythmTreeContainer(preprolated_duration=2)
        >>> c = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=3)
        >>> d = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=4)
        >>> e = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=5)

        >>> a.extend([b, c])
        >>> b.extend([d, e])

        >>> a.parentage_ratios
        (Duration(1, 1),)

        >>> b.parentage_ratios
        (Duration(1, 1), (Duration(2, 1), Duration(5, 1)))

        >>> c.parentage_ratios
        (Duration(1, 1), (Duration(3, 1), Duration(5, 1)))

        >>> d.parentage_ratios
        (Duration(1, 1), (Duration(2, 1), Duration(5, 1)), (Duration(4, 1), Duration(9, 1)))

        >>> e.parentage_ratios
        (Duration(1, 1), (Duration(2, 1), Duration(5, 1)), (Duration(5, 1), Duration(9, 1)))

        Returns tuple.
        """
        result = []
        node = self
        while node.parent is not None:
            result.append((
                node.preprolated_duration,
                node.parent._get_contents_duration(),
            ))
            node = node.parent
        result.append(node.preprolated_duration)
        return tuple(reversed(result))

    @property
    def preprolated_duration(self):
        """
        The node's preprolated_duration in pulses:

        >>> node = abjad.rhythmtrees.RhythmTreeLeaf(
        ...     preprolated_duration=1)
        >>> node.preprolated_duration
        Duration(1, 1)

        >>> node.preprolated_duration = 2
        >>> node.preprolated_duration
        Duration(2, 1)

        Returns int.
        """
        return self._duration

    @preprolated_duration.setter
    def preprolated_duration(self, argument):
        import abjad
        if not isinstance(argument, Fraction):
            argument = abjad.Duration(argument)
        assert 0 < argument
        self._duration = argument
        self._mark_entire_tree_for_later_update()

    @property
    def pretty_rtm_format(self):
        """
        The node's pretty-printed RTM format:

        >>> rtm = '(1 ((1 (1 1)) (1 (1 1))))'
        >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
        >>> print(tree.pretty_rtm_format)
        (1 (
            (1 (
                1
                1))
            (1 (
                1
                1))))

        Returns string.
        """
        return '\n'.join(self._pretty_rtm_format_pieces)

    @property
    def prolation(self):
        """
        Prolation of rhythm tree node.

        Returns multiplier.
        """
        return mathtools.cumulative_products(self.prolations)[-1]

    @property
    def prolations(self):
        """
        Prolations of rhythm tree node.

        Returns tuple.
        """
        import abjad
        prolations = [abjad.Multiplier(1)]
        pairs = abjad.sequence(self.parentage).nwise()
        for child, parent in pairs:
            prolations.append(abjad.Multiplier(
                parent.preprolated_duration, parent._get_contents_duration()))
        return tuple(prolations)

    @abc.abstractproperty
    def rtm_format(self):
        """
        The node's RTM format:

        >>> rtm = '(1 ((1 (1 1)) (1 (1 1))))'
        >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
        >>> tree.rtm_format
        '(1 ((1 (1 1)) (1 (1 1))))'

        Returns string.
        """
        raise NotImplementedError

    @property
    def start_offset(self):
        """
        The starting offset of a node in a rhythm-tree relative the root.

        >>> rtm = '(1 ((1 (1 1)) (1 (1 1))))'
        >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]

        >>> tree.start_offset
        Offset(0, 1)

        >>> tree[1].start_offset
        Offset(1, 2)

        >>> tree[0][1].start_offset
        Offset(1, 4)

        Returns Offset instance.
        """
        self._update_offsets_of_entire_tree_if_necessary()
        return self._offset

    @property
    def stop_offset(self):
        """
        The stopping offset of a node in a rhythm-tree relative the root.
        """
        return self.start_offset + self.duration


class RhythmTreeLeaf(RhythmTreeMixin, uqbar.containers.UniqueTreeNode):
    """
    Rhythm-tree leaf.

    ..  container:: example

        >>> leaf = abjad.rhythmtrees.RhythmTreeLeaf(
        ...     preprolated_duration=5, is_pitched=True)
        >>> leaf
        RhythmTreeLeaf(preprolated_duration=Duration(5, 1), is_pitched=True)

    ..  container:: example

        Calls with a pulse preprolated duration to generate leaves:

        >>> result = leaf((1, 8))
        >>> result
        Selection([Note("c'2"), Note("c'8")])

    ..  container:: example

        Generates rests when called if ``is_pitched`` is false:

        >>> abjad.rhythmtrees.RhythmTreeLeaf(
        ...     preprolated_duration=7, is_pitched=False)((1, 16))
        Selection([Rest('r4..')])

    """

    ### INITIALIZER ###

    def __init__(
        self,
        preprolated_duration=1,
        is_pitched=True,
        name=None,
    ):
        uqbar.containers.UniqueTreeNode.__init__(self, name=name)
        RhythmTreeMixin.__init__(self, preprolated_duration=preprolated_duration)
        self.is_pitched = is_pitched

    ### SPECIAL METHODS ###

    def __call__(self, pulse_duration):
        """
        Generate Abjad score components:

        >>> leaf = abjad.rhythmtrees.RhythmTreeLeaf(5)
        >>> leaf((1, 4))
        Selection([Note("c'1"), Note("c'4")])

        Returns sequence of components.
        """
        import abjad
        pulse_duration = abjad.Duration(pulse_duration)
        total_duration = pulse_duration * self.preprolated_duration
        maker = abjad.LeafMaker()
        if self.is_pitched:
            return maker(0, total_duration)
        return maker([None], total_duration)

    def __graph__(self, **keywords):
        """
        Graphviz graph of rhythm tree leaf.
        """
        graph = uqbar.graphs.Graph(name='G')
        node = uqbar.graphs.Node(
            attributes={
                'label': str(self.preprolated_duration),
                'shape': 'box'
                }
            )
        graph.append(node)
        return graph

    ### PRIVATE PROPERTIES ###

    @property
    def _pretty_rtm_format_pieces(self):
        return [str(self.preprolated_duration)]

    ### PUBLIC PROPERTIES ###

    @property
    def rtm_format(self):
        """
        RTM format of rhythm tree leaf.

        >>> abjad.rhythmtrees.RhythmTreeLeaf(1, is_pitched=True).rtm_format
        '1'
        >>> abjad.rhythmtrees.RhythmTreeLeaf(5, is_pitched=False).rtm_format
        '-5'

        Returns string.
        """
        if self.is_pitched:
            return '{!s}'.format(self.preprolated_duration)
        return '-{!s}'.format(self.preprolated_duration)

    ### PUBLIC PROPERTIES ###

    @property
    def is_pitched(self):
        """
        Gets and sets boolean equal to  true if leaf is pitched.

        >>> leaf = abjad.rhythmtrees.RhythmTreeLeaf()
        >>> leaf.is_pitched
        True

        >>> leaf.is_pitched = False
        >>> leaf.is_pitched
        False

        Returns true or false.
        """
        return self._is_pitched

    @is_pitched.setter
    def is_pitched(self, argument):
        self._is_pitched = bool(argument)


class RhythmTreeContainer(RhythmTreeMixin, uqbar.containers.UniqueTreeContainer):
    r"""
    Rhythm-tree container.

    ..  container:: example

        Initializes rhythm-tree container:

        >>> container = abjad.rhythmtrees.RhythmTreeContainer(
        ...     preprolated_duration=1,
        ...     children=[],
        ...     )

        >>> abjad.f(container)
        abjad.rhythmtrees.RhythmTreeContainer(
            children=(),
            preprolated_duration=abjad.Duration(1, 1),
            )

    ..  container:: example

        Similar to Abjad containers, ``RhythmTreeContainer`` supports a list
        interface, and can be appended, extended, indexed and so forth by other
        ``RhythmTreeMixin`` subclasses:

        >>> leaf_a = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=1)
        >>> leaf_b = abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=2)
        >>> container.extend([leaf_a, leaf_b])
        >>> abjad.f(container)
        abjad.rhythmtrees.RhythmTreeContainer(
            children=(
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    is_pitched=True,
                    ),
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(2, 1),
                    is_pitched=True,
                    ),
                ),
            preprolated_duration=abjad.Duration(1, 1),
            )

        >>> another_container = abjad.rhythmtrees.RhythmTreeContainer(
        ...     preprolated_duration=2)
        >>> another_container.append(
        ...     abjad.rhythmtrees.RhythmTreeLeaf(preprolated_duration=3))
        >>> another_container.append(container[1])
        >>> container.append(another_container)
        >>> abjad.f(container)
        abjad.rhythmtrees.RhythmTreeContainer(
            children=(
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    is_pitched=True,
                    ),
                abjad.rhythmtrees.RhythmTreeContainer(
                    children=(
                        abjad.rhythmtrees.RhythmTreeLeaf(
                            preprolated_duration=abjad.Duration(3, 1),
                            is_pitched=True,
                            ),
                        abjad.rhythmtrees.RhythmTreeLeaf(
                            preprolated_duration=abjad.Duration(2, 1),
                            is_pitched=True,
                            ),
                        ),
                    preprolated_duration=abjad.Duration(2, 1),
                    ),
                ),
            preprolated_duration=abjad.Duration(1, 1),
            )

    ..  container:: example

        Call ``RhythmTreeContainer`` with a preprolated_duration to generate a
        tuplet structure:

        >>> components = container((1, 4))
        >>> tuplet = components[0]
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \times 2/3 {
                c'8
                \times 4/5 {
                    c'8.
                    c'8
                }
            }

    Returns ``RhythmTreeContainer`` instance.
    """

    ### CLASS VARIABLES ###

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, children=None, preprolated_duration=1, name=None):
        uqbar.containers.UniqueTreeContainer.__init__(self, name=name)
        RhythmTreeMixin.__init__(self, preprolated_duration=preprolated_duration)
        if isinstance(children, (list, str, tuple)):
            self.extend(children)
        elif children is not None:
            message = 'can not instantiate {} with {!r}.'
            raise ValueError(message.format(type(self), children))

    ### SPECIAL METHODS ###

    def __add__(self, argument):
        r"""
        Concatenate containers self and argument. The operation c = a + b
        returns a new RhythmTreeContainer c with the content of both a and b,
        and a preprolated_duration equal to the sum of the durations
        of a and b. The operation is non-commutative: the content of the
        first operand will be placed before the content of the second operand:

        >>> a = abjad.rhythmtrees.RhythmTreeParser()('(1 (1 1 1))')[0]
        >>> b = abjad.rhythmtrees.RhythmTreeParser()('(2 (3 4))')[0]

        >>> c = a + b

        >>> c.preprolated_duration
        Duration(3, 1)

        >>> abjad.f(c)
        abjad.rhythmtrees.RhythmTreeContainer(
            children=(
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    is_pitched=True,
                    ),
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    is_pitched=True,
                    ),
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    is_pitched=True,
                    ),
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(3, 1),
                    is_pitched=True,
                    ),
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(4, 1),
                    is_pitched=True,
                    ),
                ),
            preprolated_duration=abjad.Duration(3, 1),
            )

        Returns new RhythmTreeContainer.
        """
        if isinstance(argument, str):
            argument = RhythmTreeParser()(argument)
            assert 1 == len(argument) and isinstance(argument[0], type(self))
            argument = argument[0]
        container = type(self)(
            preprolated_duration=self.preprolated_duration +
            argument.preprolated_duration)
        container.extend(self[:])
        container.extend(argument[:])
        return container

    def __call__(self, pulse_duration):
        r"""
        Generates Abjad score components.

        ..  container:: example

            >>> rtm = '(1 (1 (2 (1 1 1)) 2))'
            >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]

            >>> tree((1, 4))
            [Tuplet(Multiplier(4, 5), "c'16 { 2/3 c'16 c'16 c'16 } c'8")]

            >>> staff = abjad.Staff(tree((1, 4)))
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \times 4/5 {
                        c'16
                        \times 2/3 {
                            c'16
                            c'16
                            c'16
                        }
                        c'8
                    }
                }

        Returns list of components.
        """
        def recurse(node, tuplet_duration):
            basic_prolated_duration = \
                tuplet_duration / node._get_contents_duration()
            basic_written_duration = \
                basic_prolated_duration.equal_or_greater_power_of_two
            tuplet = abjad.Tuplet(1, [])
            for child in node.children:
                if isinstance(child, type(self)):
                    tuplet.extend(
                        recurse(
                            child,
                            child.preprolated_duration * basic_written_duration
                            )
                        )
                else:
                    leaves = child(basic_written_duration)
                    tuplet.extend(leaves)
                    if 1 < len(leaves):
                        tie = abjad.Tie()
                        abjad.attach(tie, leaves)
            assert tuplet.multiplier == 1, repr(tuplet.multiplier)
            contents_duration = abjad.inspect(tuplet).duration()
            target_duration = tuplet_duration
            multiplier = target_duration / contents_duration
            tuplet.multiplier = multiplier
            if tuplet.multiplier == 1:
                return tuplet[:]
            return [tuplet]
        import abjad
        pulse_duration = abjad.Duration(pulse_duration)
        assert 0 < pulse_duration
        result = recurse(self, pulse_duration * self.preprolated_duration)
        for component in result[:]:
            if isinstance(component, abjad.Tuplet):
                if component.trivial():
                    component._extract()
        return result

    def __graph__(self, **keywords):
        r"""
        The Graph representation of the RhythmTreeContainer:

        >>> rtm = '(1 (1 (2 (1 1 1)) 2))'
        >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
        >>> graph = tree.__graph__()
        >>> print(format(graph, 'graphviz'))
        digraph G {
            graph [bgcolor=transparent,
                truecolor=true];
            node_0 [label="1",
                shape=triangle];
            node_1 [label="1",
                shape=box];
            node_2 [label="2",
                shape=triangle];
            node_3 [label="1",
                shape=box];
            node_4 [label="1",
                shape=box];
            node_5 [label="1",
                shape=box];
            node_6 [label="2",
                shape=box];
            node_0 -> node_1;
            node_0 -> node_2;
            node_0 -> node_6;
            node_2 -> node_3;
            node_2 -> node_4;
            node_2 -> node_5;
        }

        >>> abjad.graph(graph) # doctest: +SKIP

        Return ``Graph`` instance.
        """
        graph = uqbar.graphs.Graph(
            name='G',
            attributes={
                'bgcolor': 'transparent',
                'truecolor': True,
                },
            )
        node_mapping = {}
        nodes = [self]
        nodes.extend(self.depth_first())
        for node in nodes:
            graphviz_node = uqbar.graphs.Node()
            graphviz_node.attributes['label'] = str(node.preprolated_duration)
            if isinstance(node, type(self)):
                graphviz_node.attributes['shape'] = 'triangle'
            else:
                graphviz_node.attributes['shape'] = 'box'
            graph.append(graphviz_node)
            node_mapping[node] = graphviz_node
            if node.parent is not None:
                uqbar.graphs.Edge().attach(
                    node_mapping[node.parent],
                    node_mapping[node],
                    )
        return graph

    def __radd__(self, argument):
        """
        Concatenates containers argument and self.

        Returns new RhythmTreeContainer.
        """
        assert isinstance(argument, type(self))
        return argument.__add__(self)

    def __repr__(self):
        """
        Gets interpreter representation of rhythm tree container.

        Returns string.
        """
        return '{}(({}, {}))'.format(
            type(self).__name__,
            self.duration.numerator,
            self.duration.denominator,
            )

    ### PRIVATE METHODS ###

    def _prepare_setitem_single(self, expr):
        if isinstance(expr, str):
            expr = RhythmTreeParser()(expr)[0]
            assert len(expr) == 1
            expr = expr[0]
        return expr

    def _prepare_setitem_multiple(self, expr):
        if isinstance(expr, str):
            expr = RhythmTreeParser()(expr)
        elif (
            isinstance(expr, list) and
            len(expr) == 1 and
            isinstance(expr[0], str)
        ):
            expr = RhythmTreeParser()(expr[0])
        return expr

    ### PRIVATE PROPERTIES ###

    def _get_contents_duration(self):
        """
        The total preprolated_duration of the children
        of a ``RhythmTreeContainer`` instance:

        >>> rtm = '(1 (1 (2 (1 1 1)) 2))'
        >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]

        >>> tree._get_contents_duration()
        Duration(5, 1)

        >>> tree[1]._get_contents_duration()
        Duration(3, 1)

        Returns int.
        """
        return sum([x.preprolated_duration for x in self])

    @property
    def _leaf_class(self):
        return RhythmTreeLeaf

    @property
    def _node_class(self):
        return RhythmTreeMixin

    @property
    def _pretty_rtm_format_pieces(self):
        result = []
        result.append('({!s} ('.format(self.preprolated_duration))
        for child in self:
            result.extend(['    ' + x for x in child._pretty_rtm_format_pieces])
        result[-1] = result[-1] + '))'
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def rtm_format(self):
        """
        The node's RTM format:

        >>> rtm = '(1 ((1 (1 1)) (1 (1 1))))'
        >>> tree = abjad.rhythmtrees.RhythmTreeParser()(rtm)[0]
        >>> tree.rtm_format
        '(1 ((1 (1 1)) (1 (1 1))))'

        Returns string.
        """
        return '({!s} ({}))'.format(
            self.preprolated_duration,
            ' '.join([x.rtm_format for x in self]))


class RhythmTreeParser(Parser):
    r"""
    Rhythm-tree parser.

    ..  container:: example

        Abjad’s rhythm-tree parser parses a micro-language resembling Ircam’s
        RTM Lisp syntax, and generates a sequence of RhythmTree structures,
        which can be furthered manipulated by composers, before being converted
        into an Abjad score object:

        >>> parser = abjad.rhythmtrees.RhythmTreeParser()

        >>> string = '(3 (1 (1 ((2 (1 1 1)) 2 2 1))))'
        >>> rhythm_tree_list = parser(string)
        >>> rhythm_tree_container = rhythm_tree_list[0]
        >>> rhythm_tree_container.rtm_format
        '(3 (1 (1 ((2 (1 1 1)) 2 2 1))))'

        >>> abjad.f(rhythm_tree_container)
        abjad.rhythmtrees.RhythmTreeContainer(
            children=(
                abjad.rhythmtrees.RhythmTreeLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    is_pitched=True,
                    ),
                abjad.rhythmtrees.RhythmTreeContainer(
                    children=(
                        abjad.rhythmtrees.RhythmTreeContainer(
                            children=(
                                abjad.rhythmtrees.RhythmTreeLeaf(
                                    preprolated_duration=abjad.Duration(1, 1),
                                    is_pitched=True,
                                    ),
                                abjad.rhythmtrees.RhythmTreeLeaf(
                                    preprolated_duration=abjad.Duration(1, 1),
                                    is_pitched=True,
                                    ),
                                abjad.rhythmtrees.RhythmTreeLeaf(
                                    preprolated_duration=abjad.Duration(1, 1),
                                    is_pitched=True,
                                    ),
                                ),
                            preprolated_duration=abjad.Duration(2, 1),
                            ),
                        abjad.rhythmtrees.RhythmTreeLeaf(
                            preprolated_duration=abjad.Duration(2, 1),
                            is_pitched=True,
                            ),
                        abjad.rhythmtrees.RhythmTreeLeaf(
                            preprolated_duration=abjad.Duration(2, 1),
                            is_pitched=True,
                            ), abjad.rhythmtrees.RhythmTreeLeaf(
                            preprolated_duration=abjad.Duration(1, 1),
                            is_pitched=True,
                            ),
                        ),
                    preprolated_duration=abjad.Duration(1, 1),
                    ),
                ),
            preprolated_duration=abjad.Duration(3, 1),
            )

        >>> base_duration = (1, 4)
        >>> component_list = rhythm_tree_container(base_duration)
        >>> tuplet = component_list[0]
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \tweak text #tuplet-number::calc-fraction-text
            \times 3/4 {
                c'2
                \times 4/7 {
                    \times 2/3 {
                        c'8
                        c'8
                        c'8
                    }
                    c'4
                    c'4
                    c'8
                }
            }

    """

    ### PUBLIC PROPERTIES ###

    @property
    def lexer_rules_object(self):
        return self

    @property
    def parser_rules_object(self):
        return self

    ### LEX SETUP ###

    tokens = (
        'DURATION',
        'LPAREN',
        'RPAREN'
    )

    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_ignore = ' \n\t\r'

    ### YACC SETUP ###

    start = 'toplevel'

    ### LEX METHODS ###

    def t_DURATION(self, t):
        r'-?[1-9]\d*(/[1-9]\d*)?'
        import abjad
        parts = t.value.partition('/')
        if not parts[2]:
            t.value = abjad.Duration(int(parts[0]))
        else:
            numerator, denominator = int(parts[0]), int(parts[2])
            fraction = abjad.NonreducedFraction(numerator, denominator)
            preprolated_duration = abjad.Duration(fraction)
            if fraction.numerator == preprolated_duration.numerator:
                t.value = preprolated_duration
            else:
                t.value = fraction
        return t

    def t_error(self, t):
        print(("Illegal character '%s'" % t.value[0]))
        t.lexer.skip(1)

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")

    ### YACC METHODS ###

    def p_container__LPAREN__DURATION__node_list_closed__RPAREN(self, p):
        """
        container : LPAREN DURATION node_list_closed RPAREN
        """
        import abjad
        p[0] = abjad.rhythmtrees.RhythmTreeContainer(
            children=p[3],
            preprolated_duration=abs(p[2]),
            )

    def p_error(self, p):
        if p:
            print(("Syntax error at '%s'" % p.value))
        else:
            print("Syntax error at EOF")

    def p_leaf__INTEGER(self, p):
        """
        leaf : DURATION
        """
        import abjad
        p[0] = abjad.rhythmtrees.RhythmTreeLeaf(
            preprolated_duration=abs(p[1]),
            is_pitched=0 < p[1],
            )

    def p_node__container(self, p):
        """
        node : container
        """
        p[0] = p[1]

    def p_node__leaf(self, p):
        """
        node : leaf
        """
        p[0] = p[1]

    def p_node_list__node_list__node_list_item(self, p):
        """
        node_list : node_list node_list_item
        """
        p[0] = p[1] + [p[2]]

    def p_node_list__node_list_item(self, p):
        """
        node_list : node_list_item
        """
        p[0] = [p[1]]

    def p_node_list_closed__LPAREN__node_list__RPAREN(self, p):
        """
        node_list_closed : LPAREN node_list RPAREN
        """
        p[0] = p[2]

    def p_node_list_item__node(self, p):
        """
        node_list_item : node
        """
        p[0] = p[1]

    def p_toplevel__EMPTY(self, p):
        """
        toplevel :
        """
        p[0] = []

    def p_toplevel__toplevel__node(self, p):
        """
        toplevel : toplevel node
        """
        p[0] = p[1] + [p[2]]


def parse_rtm_syntax(rtm):
    r"""
    Parses RTM syntax.

    ..  container:: example

        Parses tuplet:

        >>> rtm = '(1 (1 (1 (1 1)) 1))'
        >>> tuplet = abjad.rhythmtrees.parse_rtm_syntax(rtm)
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \times 2/3 {
                c'8
                c'16
                c'16
                c'8
            }

    ..  container:: example

        Also supports fractional durations:

        >>> rtm = '(3/4 (1 1/2 (4/3 (1 -1/2 1))))'
        >>> tuplet = abjad.rhythmtrees.parse_rtm_syntax(rtm)
        >>> abjad.show(tuplet) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(tuplet)
            \tweak text #tuplet-number::calc-fraction-text
            \times 9/17 {
                c'8
                c'16
                \tweak edge-height #'(0.7 . 0)
                \times 8/15 {
                    c'8
                    r16
                    c'8
                }
            }

    Returns tuplet or container.
    """
    result = RhythmTreeParser()(rtm)
    container = core.Container()
    for node in result:
        tuplet = node((1, 4))
        # following line added 2012-08-01. tb.
        tuplet = tuplet[0]
        if tuplet.trivial():
            container.extend(tuplet[:])
        else:
            container.append(tuplet)
    if len(container) == 1:
        return container[0]
    return container



#! /usr/bin/env python
import abjad
import copy

__documentation_section__ = 'demos'


def make_bartok_score():
    """
    Build the Bartok example score.
    """

    # Build score skeleton
    score = abjad.Score()
    piano_staff = abjad.StaffGroup(lilypond_type='PianoStaff')
    upper_staff = abjad.Staff([])
    lower_staff = abjad.Staff([])
    piano_staff.append(upper_staff)
    piano_staff.append(lower_staff)
    score.append(piano_staff)

    # Build upper measures
    upper_measures = []
    upper_measures.append(abjad.Measure((2, 4), []))
    upper_measures.append(abjad.Measure((3, 4), []))
    upper_measures.append(abjad.Measure((2, 4), []))
    upper_measures.append(abjad.Measure((2, 4), []))
    upper_measures.append(abjad.Measure((2, 4), []))
    lower_measures = copy.deepcopy(upper_measures)
    upper_staff.extend(upper_measures)
    lower_staff.extend(lower_measures)

    # Add leaves to upper measures
    upper_measures[0].extend("a'8 g'8 f'8 e'8")
    upper_measures[1].extend("d'4 g'8 f'8 e'8 d'8")
    upper_measures[2].extend("c'8 d'16 e'16 f'8 e'8")
    upper_measures[3].append("d'2")
    upper_measures[4].append("d'2")

    # Add leaves to lower measures
    lower_measures[0].extend("b4 d'8 c'8")
    lower_measures[1].extend("b8 a8 af4 c'8 bf8")
    lower_measures[2].extend("a8 g8 fs8 g16 a16")

    # Build parallel music for measure 4
    upper_voice = abjad.Voice("b2", name='upper voice')
    command = abjad.LilyPondLiteral(r'\voiceOne')
    abjad.attach(command, upper_voice)
    lower_voice = abjad.Voice("b4 a4", name='lower voice')
    command = abjad.LilyPondLiteral(r'\voiceTwo')
    abjad.attach(command, lower_voice)
    lower_measures[3].extend([upper_voice, lower_voice])
    lower_measures[3].is_simultaneous = True

    # Build parallel music for measure 5
    upper_voice = abjad.Voice("b2", name='upper voice')
    command = abjad.LilyPondLiteral(r'\voiceOne')
    abjad.attach(command, upper_voice)
    lower_voice = abjad.Voice("g2", name='lower voice')
    command = abjad.LilyPondLiteral(r'\voiceTwo')
    abjad.attach(command, lower_voice)
    lower_measures[4].extend([upper_voice, lower_voice])
    lower_measures[4].is_simultaneous = True

    # Add bass clef
    clef = abjad.Clef('bass')
    leaf = abjad.inspect(lower_staff).leaf(0)
    abjad.attach(clef, leaf)

    # Add dynamics
    dynamic = abjad.Dynamic('pp')
    abjad.attach(dynamic, upper_measures[0][0])
    dynamic = abjad.Dynamic('mp')
    abjad.attach(dynamic, upper_measures[1][1])
    dynamic = abjad.Dynamic('pp')
    abjad.attach(dynamic, lower_measures[0][1])
    dynamic = abjad.Dynamic('mp')
    abjad.attach(dynamic, lower_measures[1][3])

    # Add final bar line
    score.add_final_bar_line()

    # Select leaves for attaching spanners to
    upper_leaves = abjad.select(upper_staff).leaves()
    lower_leaves = abjad.select(lower_staff).leaves()

    # Attach beams
    beam = abjad.Beam()
    abjad.attach(beam, upper_leaves[:4])
    beam = abjad.Beam()
    abjad.attach(beam, lower_leaves[1:5])
    beam = abjad.Beam()
    abjad.attach(beam, lower_leaves[6:10])

    # Attach slurs
    slur = abjad.Slur()
    abjad.attach(slur, upper_leaves[:5])
    slur = abjad.Slur()
    abjad.attach(slur, upper_leaves[5:])
    slur = abjad.Slur()
    abjad.attach(slur, lower_leaves[1:6])

    # Attach hairpins
    crescendo = abjad.Hairpin('<')
    abjad.attach(crescendo, upper_leaves[-7:-2])
    decrescendo = abjad.Hairpin('>')
    abjad.attach(decrescendo, upper_leaves[-2:])

    # Attach a ritardando with markup
    markup = abjad.Markup('ritard.')
    text_spanner = abjad.TextSpanner()
    abjad.tweak(text_spanner).bound_details__left__text = markup
    abjad.attach(text_spanner, upper_leaves[-7:])

    # Tie notes
    tie = abjad.Tie()
    abjad.attach(tie, upper_leaves[-2:])

    # Tie more notes
    note_1 = lower_staff[-2]['upper voice'][0]
    note_2 = lower_staff[-1]['upper voice'][0]
    notes = abjad.select([note_1, note_2])
    abjad.attach(abjad.Tie(), notes)

    # Return the score
    return score


if __name__ == '__main__':
    # for mypy
    from abjad import show

    lilypond_file = make_bartok_score()
    show(lilypond_file)



import typing
from abjad.system.AbjadObject import AbjadObject


class CyclicTuple(AbjadObject):
    """
    Cyclic tuple.

    ..  container:: example

        Initializes from string:

        >>> tuple_ = abjad.CyclicTuple('abcd')

        >>> tuple_
        CyclicTuple(['a', 'b', 'c', 'd'])

        >>> for x in range(8):
        ...     print(x, tuple_[x])
        ...
        0 a
        1 b
        2 c
        3 d
        4 a
        5 b
        6 c
        7 d

    Cyclic tuples overload the item-getting method of built-in tuples.

    Cyclic tuples return a value for any integer index.

    Cyclic tuples otherwise behave exactly like built-in tuples.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_items',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        items: typing.Sequence = None,
        ) -> None:
        items = items or ()
        items = tuple(items)
        self._items: typing.Tuple = items

    ### SPECIAL METHODS ###

    def __contains__(self, item) -> bool:
        """
        Is true when cyclic tuple contains ``item``.
        """
        return self._items.__contains__(item)

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is a tuple with items equal to those of this
        cyclic tuple.
        """
        if isinstance(argument, tuple):
            return self._items == argument
        elif isinstance(argument, type(self)):
            return self._items == argument._items
        return False

    def __getitem__(self, argument) -> typing.Any:
        """
        Gets item or slice identified by ``argument``.

        ..  container:: example

            Gets slice open at right:

            >>> items = [0, 1, 2, 3, 4, 5]
            >>> tuple_ = abjad.CyclicTuple(items=items)
            >>> tuple_[2:]
            (2, 3, 4, 5)

        ..  container:: example

            Gets slice closed at right:

            >>> items = [0, 1, 2, 3, 4, 5]
            >>> tuple_ = abjad.CyclicTuple(items=items)
            >>> tuple_[:15]
            (0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2)

        Raises index error when ``argument`` can not be found in cyclic tuple.
        """
        if isinstance(argument, slice):
            if ((argument.stop is not None and argument.stop < 0) or
                (argument.start is not None and argument.start < 0)):
                return self._items.__getitem__(argument)
            else:
                return self._get_slice(argument.start, argument.stop)
        if not self:
            raise IndexError(f'cyclic tuple is empty: {self!r}.')
        argument = argument % len(self)
        return self._items.__getitem__(argument)

    def __hash__(self) -> int:
        """
        Hashes cyclic tuple.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    def __iter__(self) -> typing.Iterator:
        """
        Iterates cyclic tuple.

        Iterates items only once.

        Does not iterate infinitely.
        """
        return self._items.__iter__()

    def __len__(self) -> int:
        """
        Gets length of cyclic tuple.
        """
        assert isinstance(self._items, tuple)
        return self._items.__len__()

    def __str__(self) -> str:
        """
        Gets string representation of cyclic tuple.

        ..  container:: example

            Gets string:

            >>> str(abjad.CyclicTuple('abcd'))
            '(a, b, c, d)'

        ..  container:: example

            Gets string:

            >>> str(abjad.CyclicTuple([1, 2, 3, 4]))
            '(1, 2, 3, 4)'

        """
        if self:
            contents = [str(item) for item in self._items]
            string = ', '.join(contents)
            string = f'({string})'
            return string
        return '()'

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        import abjad
        return abjad.FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_args_values=[list(self._items)],
            )

    def _get_slice(self, start_index, stop_index):
        if stop_index is not None and 1000000 < stop_index:
            stop_index = len(self)
        result = []
        if start_index is None:
            start_index = 0
        if stop_index is None:
            indices = range(start_index, len(self))
        else:
            indices = range(start_index, stop_index)
        result = [self[n] for n in indices]
        return tuple(result)

    ### PUBLIC PROPERTIES ###

    @property
    def items(self) -> typing.Tuple:
        """
        Gets items in cyclic tuple.

        ..  container:: example

            Gets items:

            >>> tuple_ = abjad.CyclicTuple('abcd')
            >>> tuple_.items
            ('a', 'b', 'c', 'd')

        ..  container:: example

            Gets items:

            >>> tuple_ = abjad.CyclicTuple([1, 2, 3, 4])
            >>> tuple_.items
            (1, 2, 3, 4)

        """
        return self._items



import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.utilities.String import String
from abjad.system.FormatSpecification import FormatSpecification
from .Part import Part


class PartAssignment(AbjadValueObject):
    """
    Part assignment.

    ..  container:: example

        >>> abjad.PartAssignment('Horn')
        PartAssignment('Horn')

        >>> abjad.PartAssignment('Horn', 1)
        PartAssignment('Horn', 1)

        >>> abjad.PartAssignment('Horn', 2)
        PartAssignment('Horn', 2)

        >>> abjad.PartAssignment('Horn', (3, 4))
        PartAssignment('Horn', (3, 4))

        >>> abjad.PartAssignment('Horn', [1, 3])
        PartAssignment('Horn', [1, 3])

    ..  container:: example

        >>> part_assignment = abjad.PartAssignment('Horn', [1, 3])

        >>> abjad.f(part_assignment)
        PartAssignment('Horn', [1, 3])

        >>> print(format(part_assignment))
        abjad.PartAssignment('Horn', [1, 3])
        
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_members',
        '_parts',
        '_section',
        '_token',
        )

    ### INITIALIZER ###

    token_type = typing.Union[
        None,
        int,
        typing.Tuple[int, int],
        typing.List[int],
        ]

    def __init__(
        self,
        section: str = None,
        token: token_type = None,
        ) -> None:
        self._section = section
        if token is not None:
            assert self._is_token(token), repr(token)
        self._token = token
        members = self._expand_members(token)
        self._members = members
        parts = self._expand_parts()
        assert isinstance(parts, list), repr(parts)
        self._parts = parts

    ### SPECIAL METHODS ###

    def __contains__(self, part: Part) -> bool:
        """
        Is true when part assignment contains ``part``.

        ..  container:: example

            >>> parts = [
            ...     abjad.Part(section='Horn', member= 1),
            ...     abjad.Part(section='Horn', member= 2),
            ...     abjad.Part(section='Horn', member= 3),
            ...     abjad.Part(section='Horn', member= 4),
            ...     ]

            >>> part_assignment = abjad.PartAssignment('Horn')
            >>> for part in parts:
            ...     part, part in part_assignment
            ...
            (Part(instrument='Horn', member=1, section='Horn'), True)
            (Part(instrument='Horn', member=2, section='Horn'), True)
            (Part(instrument='Horn', member=3, section='Horn'), True)
            (Part(instrument='Horn', member=4, section='Horn'), True)

            >>> part_assignment = abjad.PartAssignment('Horn', 1)
            >>> for part in parts:
            ...     part, part in part_assignment
            ...
            (Part(instrument='Horn', member=1, section='Horn'), True)
            (Part(instrument='Horn', member=2, section='Horn'), False)
            (Part(instrument='Horn', member=3, section='Horn'), False)
            (Part(instrument='Horn', member=4, section='Horn'), False)

            >>> part_assignment = abjad.PartAssignment('Horn', 2)
            >>> for part in parts:
            ...     part, part in part_assignment
            ...
            (Part(instrument='Horn', member=1, section='Horn'), False)
            (Part(instrument='Horn', member=2, section='Horn'), True)
            (Part(instrument='Horn', member=3, section='Horn'), False)
            (Part(instrument='Horn', member=4, section='Horn'), False)

            >>> part_assignment = abjad.PartAssignment('Horn', (3, 4))
            >>> for part in parts:
            ...     part, part in part_assignment
            ...
            (Part(instrument='Horn', member=1, section='Horn'), False)
            (Part(instrument='Horn', member=2, section='Horn'), False)
            (Part(instrument='Horn', member=3, section='Horn'), True)
            (Part(instrument='Horn', member=4, section='Horn'), True)

            >>> part_assignment = abjad.PartAssignment('Horn', [1, 3])
            >>> for part in parts:
            ...     part, part in part_assignment
            ...
            (Part(instrument='Horn', member=1, section='Horn'), True)
            (Part(instrument='Horn', member=2, section='Horn'), False)
            (Part(instrument='Horn', member=3, section='Horn'), True)
            (Part(instrument='Horn', member=4, section='Horn'), False)

        ..  container:: example

            Raises exception when input is not part:

            >>> part_assignment = abjad.PartAssignment('Horn')
            >>> 'Horn' in part_assignment
            Traceback (most recent call last):
                ...
            TypeError: must be part (not 'Horn').

        """
        if not isinstance(part, Part):
            raise TypeError(f'must be part (not {part!r}).')
        if part.section == self.section:
            if (part.member is None or
                self.members is None or
                part.member in self.members or []):
                return True
            return False
        return False

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is a part assignment with section and
        members equal to this part assignment.

        ..  container:: example

            >>> part_assignment_1 = abjad.PartAssignment('Horn', (1, 2))
            >>> part_assignment_2 = abjad.PartAssignment('Horn', [1, 2])
            >>> part_assignment_3 = abjad.PartAssignment('Horn')

            >>> part_assignment_1 == part_assignment_1
            True
            >>> part_assignment_1 == part_assignment_2
            True
            >>> part_assignment_1 == part_assignment_3
            False

            >>> part_assignment_2 == part_assignment_1
            True
            >>> part_assignment_2 == part_assignment_2
            True
            >>> part_assignment_2 == part_assignment_3
            False

            >>> part_assignment_3 == part_assignment_1
            False
            >>> part_assignment_3 == part_assignment_2
            False
            >>> part_assignment_3 == part_assignment_3
            True

        """
        if isinstance(argument, type(self)):
            if argument.section == self.section:
                return argument.members == self.members
        return False

    def __hash__(self) -> int:
        """
        Hashes part assignment.
        """
        return super().__hash__()

    def __iter__(self):
        """
        Iterates parts in assignment.
        
        ..  container:: example

            >>> part_assignment = abjad.PartAssignment('Horn', [1, 3])
            >>> for part in part_assignment:
            ...     part
            ... 
            Part(instrument='Horn', member=1, section='Horn')
            Part(instrument='Horn', member=3, section='Horn')

        """
        return iter(self.parts)

    ### PRIVATE METHODS ###

    @staticmethod
    def _expand_members(token):
        if token is None:
            return
        members = []
        if isinstance(token, int):
            members.append(token)
        elif isinstance(token, tuple):
            assert len(token) == 2, repr(token)
            for member in range(token[0], token[1] + 1):
                members.append(member)
        else:
            assert isinstance(token, list), repr(token)
            members.extend(token)
        return members

    def _expand_parts(self):
        parts = []
        if self.members is None:
            parts.append(Part(section=self.section))
        else:
            for member in self.members:
                part = Part(member=member, section=self.section)
                parts.append(part)
        return parts

    def _get_format_specification(self):
        repr_args_values = [self.section]
        if self.token is not None:
            repr_args_values.append(self.token)
        repr_is_indented = False
        repr_kwargs_names = []
        return FormatSpecification(
            self,
            repr_args_values=repr_args_values,
            repr_is_indented=repr_is_indented,
            repr_kwargs_names=repr_kwargs_names,
            storage_format_args_values=repr_args_values,
            storage_format_is_indented=repr_is_indented,
            storage_format_kwargs_names=repr_kwargs_names,
            )

    @staticmethod
    def _is_token(argument):
        if isinstance(argument, int) and 1 <= argument:
            return True
        if (isinstance(argument, tuple) and
            len(argument) == 2 and
            isinstance(argument[0], int) and
            isinstance(argument[1], int)):
            return True
        if isinstance(argument, list):
            for item in argument:
                if not isinstance(item, int):
                    return False
                if not 1 <= item:
                    return False
            return True
        return False

    ### PUBLIC PROPERTIES ###

    @property
    def members(self) -> typing.Optional[typing.List[int]]:
        """
        Gets members.

        ..  container:: example

            >>> abjad.PartAssignment('Horn').members is None
            True

            >>> abjad.PartAssignment('Horn', 1).members
            [1]

            >>> abjad.PartAssignment('Horn', 2).members
            [2]

            >>> abjad.PartAssignment('Horn', (3, 4)).members
            [3, 4]

            >>> abjad.PartAssignment('Horn', [1, 3]).members
            [1, 3]

        """
        return self._members

    @property
    def parts(self) -> typing.List[Part]:
        """
        Gets parts.

        ..  container:: example

            >>> abjad.PartAssignment('Horn').parts
            [Part(instrument='Horn', section='Horn')]

            >>> abjad.PartAssignment('Horn', 1).parts
            [Part(instrument='Horn', member=1, section='Horn')]

            >>> abjad.PartAssignment('Horn', 2).parts
            [Part(instrument='Horn', member=2, section='Horn')]

            >>> abjad.PartAssignment('Horn', (3, 4)).parts
            [Part(instrument='Horn', member=3, section='Horn'), Part(instrument='Horn', member=4, section='Horn')]

            >>> abjad.PartAssignment('Horn', [1, 3]).parts
            [Part(instrument='Horn', member=1, section='Horn'), Part(instrument='Horn', member=3, section='Horn')]

        """
        return self._parts

    @property
    def section(self) -> typing.Optional[str]:
        """
        Gets section.

        ..  container:: example

            >>> abjad.PartAssignment('Horn').section
            'Horn'

            >>> abjad.PartAssignment('Horn', 1).section
            'Horn'

            >>> abjad.PartAssignment('Horn', 2).section
            'Horn'

            >>> abjad.PartAssignment('Horn', (3, 4)).section
            'Horn'

            >>> abjad.PartAssignment('Horn', [1, 3]).section
            'Horn'

        """
        return self._section

    @property
    def token(self) -> token_type:
        """
        Gets token.

        ..  container:: example

            >>> abjad.PartAssignment('Horn').token is None
            True

            >>> abjad.PartAssignment('Horn', 1).token
            1

            >>> abjad.PartAssignment('Horn', 2).token
            2

            >>> abjad.PartAssignment('Horn', (3, 4)).token
            (3, 4)

            >>> abjad.PartAssignment('Horn', [1, 3]).token
            [1, 3]

        """
        return self._token



import abc
import collections
import types
from abjad.system.FormatSpecification import FormatSpecification
from abjad.utilities.TypedCollection import TypedCollection
from abjad.utilities.TypedFrozenset import TypedFrozenset


class Set(TypedFrozenset):
    """
    Abstract set.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_expression',
        )

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        if isinstance(items, str):
            items = items.split()
        elif isinstance(items, (
            collections.Iterator,
            types.GeneratorType,
            )):
            items = [item for item in items]
        if item_class is None:
            item_class = self._named_item_class
            if items is not None:
                if (isinstance(items, TypedCollection) and
                    issubclass(items.item_class, self._parent_item_class)):
                    item_class = items.item_class
                elif len(items):
                    if isinstance(items, collections.Set):
                        items = tuple(items)
                    if isinstance(items[0], str):
                        item_class = self._named_item_class
                    elif isinstance(items[0], (int, float)):
                        item_class = self._numbered_item_class
                    elif isinstance(items[0], self._parent_item_class):
                        item_class = type(items[0])
        assert issubclass(item_class, self._parent_item_class)
        TypedFrozenset.__init__(
            self,
            items=items,
            item_class=item_class,
            )
        self._expression = None

    ### SPECIAL METHODS ###

    def __str__(self):
        """
        Gets string representation.

        Returns string.
        """
        items = self._get_sorted_repr_items()
        items = [str(_) for _ in items]
        return '{{{}}}'.format(', '.join(items))

    ### PRIVATE PROPERTIES ###

    @abc.abstractproperty
    def _named_item_class(self):
        raise NotImplementedError

    @abc.abstractproperty
    def _numbered_item_class(self):
        raise NotImplementedError

    @abc.abstractproperty
    def _parent_item_class(self):
        raise NotImplementedError

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        repr_items = self._get_sorted_repr_items()
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            repr_args_values=[repr_items],
            storage_format_args_values=[repr_items],
            storage_format_kwargs_names=[],
            )

    def _get_sorted_repr_items(self):
        items = sorted(self, key=lambda x: (float(x.number), str(x)))
        if self.item_class.__name__.startswith('Named'):
            repr_items = [str(x) for x in items]
        elif hasattr(self.item_class, 'number'):
            repr_items = [x.number for x in items]
        elif hasattr(self.item_class, 'pitch_class_number'):
            repr_items = [x.pitch_class_number for x in items]
        elif hasattr(self.item_class, '__abs__'):
            repr_items = [abs(x) for x in items]
        else:
            message = 'invalid item class: {!r}.'
            message = message.format(self.item_class)
            raise ValueError(message)
        return repr_items

    def _sort_self(self):
        return tuple(self)

    ### PUBLIC PROPERTIES ###

    @property
    def cardinality(self):
        """
        Gets cardinality of set.

        Defined equal to length of set.

        Returns nonnegative integer.
        """
        return len(self)

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def from_selection(
        class_,
        selection,
        item_class=None,
        ):
        """
        Makes set from `selection`.

        Returns set.
        """
        raise NotImplementedError



import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle


class StopTextSpan(AbjadValueObject):
    r"""
    LilyPond ``\stopTextSpan`` command.

    ..  container:: example

        >>> abjad.StopTextSpan()
        StopTextSpan(command='\\stopTextSpan')

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_command',
        '_leak',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(
        self,
        command: str = r'\stopTextSpan',
        *,
        leak: bool = None,
        ) -> None:
        assert isinstance(command, str), repr(command)
        assert command.startswith('\\'), repr(command)
        self._command = command
        if leak is not None:
            leak = bool(leak)
        self._leak = leak

    ### PRIVATE METHODS ###

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        string = self.command
        if self.leak:
            string = f'<> {string}'
            bundle.after.leaks.append(string)
        else:
            bundle.after.spanner_stops.append(string)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def command(self) -> str:
        """
        Gets command.
        """
        return self._command

    @property
    def enchained(self) -> bool:
        """
        Is true.
        """
        return True

    @property
    def leak(self) -> typing.Optional[bool]:
        r"""
        Is true when stop text span leaks LilyPond ``<>`` empty chord.

        ..  container:: example

            Without leak:

            >>> staff = abjad.Staff("c'4 d' e' r")
            >>> command = abjad.StartTextSpan(
            ...     left_text=abjad.Markup('pont.').upright(),
            ...     right_text=abjad.Markup('tasto').upright(),
            ...     style='dashed_line_with_arrow',
            ...     )
            >>> abjad.tweak(command).staff_padding = 2.5
            >>> abjad.attach(command, staff[0])
            >>> command = abjad.StopTextSpan()
            >>> abjad.attach(command, staff[-2])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    - \abjad_dashed_line_with_arrow
                    - \tweak bound-details.left.text \markup {
                        \concat
                            {
                                \upright
                                    pont.
                                \hspace
                                    #0.5
                            }
                        }
                    - \tweak bound-details.right.text \markup {
                        \upright
                            tasto
                        }
                    - \tweak staff-padding #2.5
                    \startTextSpan
                    d'4
                    e'4
                    \stopTextSpan
                    r4
                }

            With leak:

            >>> staff = abjad.Staff("c'4 d' e' r")
            >>> command = abjad.StartTextSpan(
            ...     left_text=abjad.Markup('pont.').upright(),
            ...     right_text=abjad.Markup('tasto').upright(),
            ...     style='dashed_line_with_arrow',
            ...     )
            >>> abjad.tweak(command).staff_padding = 2.5
            >>> abjad.attach(command, staff[0])
            >>> command = abjad.StopTextSpan(leak=True)
            >>> abjad.attach(command, staff[-2])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'4
                    - \abjad_dashed_line_with_arrow
                    - \tweak bound-details.left.text \markup {
                        \concat
                            {
                                \upright
                                    pont.
                                \hspace
                                    #0.5
                            }
                        }
                    - \tweak bound-details.right.text \markup {
                        \upright
                            tasto
                        }
                    - \tweak staff-padding #2.5
                    \startTextSpan
                    d'4
                    e'4
                    <> \stopTextSpan
                    r4
                }

        ..  container:: example

            REGRESSION. Leaked contributions appear last in postevent format
            slot:

            >>> staff = abjad.Staff("c'8 d' e' f' r2")
            >>> abjad.attach(abjad.Beam(), staff[:4])
            >>> abjad.attach(abjad.Slur(), staff[:4])
            >>> command = abjad.StartTextSpan(
            ...     left_text=abjad.Markup('pont.').upright(),
            ...     right_text=abjad.Markup('tasto').upright(),
            ...     style='dashed_line_with_arrow',
            ...     )
            >>> abjad.tweak(command).staff_padding = 2.5
            >>> abjad.attach(command, staff[0])
            >>> command = abjad.StopTextSpan(leak=True)
            >>> abjad.attach(command, staff[3])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    - \abjad_dashed_line_with_arrow
                    - \tweak bound-details.left.text \markup {
                        \concat
                            {
                                \upright
                                    pont.
                                \hspace
                                    #0.5
                            }
                        }
                    - \tweak bound-details.right.text \markup {
                        \upright
                            tasto
                        }
                    - \tweak staff-padding #2.5
                    \startTextSpan
                    [
                    (
                    d'8
                    e'8
                    f'8
                    ]
                    )
                    <> \stopTextSpan
                    r2
                }

            The leaked text spanner above does not inadvertantly leak the beam
            or slur.

        """
        return self._leak

    @property
    def spanner_stop(self) -> bool:
        """
        Is true.
        """
        return True



def all_are_nonnegative_integer_equivalent_numbers(argument):
    """
    Is true when ``argument`` is an iterable collection of nonnegative
    integer-equivalent numbers.

    ..  container:: example

        >>> items = [0, 0.0, abjad.Fraction(0), 2, 2.0, abjad.Fraction(2)]
        >>> abjad.mathtools.all_are_nonnegative_integer_equivalent_numbers(items)
        True

        >>> items = [0, 0.0, abjad.Fraction(0), -2, 2.0, abjad.Fraction(2)]
        >>> abjad.mathtools.all_are_nonnegative_integer_equivalent_numbers(items)
        False

    Returns true or false.
    """
    from abjad import mathtools
    try:
        return all(
            mathtools.is_nonnegative_integer_equivalent_number(_)
            for _ in argument
            )
    except TypeError:
        return False



import typing
from abjad.lilypondnames.LilyPondContextSetting import LilyPondContextSetting
from abjad.scheme import SchemeSymbol
from .Spanner import Spanner


class PianoPedalSpanner(Spanner):
    r"""
    Piano pedal spanner.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> spanner = abjad.PianoPedalSpanner()
        >>> abjad.tweak(spanner).color = 'blue'
        >>> abjad.tweak(spanner).sustain_pedal_line_spanner.staff_padding = 5
        >>> abjad.attach(spanner, staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \set Staff.pedalSustainStyle = #'mixed
                c'8
                - \tweak SustainPedalLineSpanner.staff-padding #5
                - \tweak color #blue
                \sustainOn
                d'8
                e'8
                f'8
                \sustainOff
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_kind',
        '_style',
        )

    _kinds = {
        'sustain': (r'\sustainOn', r'\sustainOff'),
        'sostenuto': (r'\sostenutoOn', r'\sostenutoOff'),
        'corda': (r'\unaCorda', r'\treCorde'),
        }

    _styles = (
        'text',
        'bracket',
        'mixed',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        kind: str = 'sustain',
        style: str = 'mixed',
        ) -> None:
        Spanner.__init__(self)
        if kind not in list(self._kinds.keys()):
            raise ValueError(f'kind must be in {list(self._kinds.keys())!r}.')
        self._kind = kind
        if style not in self._styles:
            raise ValueError(f'style must be in {self._styles!r}.')
        self._style = style

    ### PRIVATE PROPERTIES ##

    @property
    def _start_command(self):
        return self._kinds[self.kind][0]

    @property
    def _stop_command(self):
        return self._kinds[self.kind][1]

    ### PRIVATE METHODS ###

    def _copy_keywords(self, new):
        new._kind = self.kind
        new._style = self.style

    def _get_lilypond_format_bundle(self, leaf):
        bundle = self._get_basic_lilypond_format_bundle(leaf)
        if self._is_my_only(leaf):
            style = SchemeSymbol(self.style)
            context_setting = LilyPondContextSetting(
                lilypond_type='Staff',
                context_property='pedalSustainStyle',
                value=style,
                )
            bundle.update(context_setting)
            strings = self._tweaked_start_command_strings()
            bundle.after.spanner_starts.extend(strings)
            string = self._stop_command_string()
            bundle.after.spanner_starts.append(string)
        elif leaf is self[0]:
            style = SchemeSymbol(self.style)
            context_setting = LilyPondContextSetting(
                lilypond_type='Staff',
                context_property='pedalSustainStyle',
                value=style,
                )
            bundle.update(context_setting)
            strings = self._tweaked_start_command_strings()
            bundle.after.spanner_starts.extend(strings)
        elif leaf is self[-1]:
            string = self._stop_command_string()
            bundle.after.spanner_stops.append(string)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def kind(self) -> str:
        r"""
        Gets kind of piano pedal spanner.

        ..  container:: example

            Sustain pedal:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> spanner = abjad.PianoPedalSpanner(kind='sustain')
            >>> abjad.tweak(spanner).color = 'blue'
            >>> abjad.tweak(spanner).sustain_pedal_line_spanner.staff_padding = 5
            >>> abjad.attach(spanner, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \set Staff.pedalSustainStyle = #'mixed
                    c'8
                    - \tweak SustainPedalLineSpanner.staff-padding #5
                    - \tweak color #blue
                    \sustainOn
                    d'8
                    e'8
                    f'8
                    \sustainOff
                }

        ..  container:: example

            Sostenuto pedal:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> spanner = abjad.PianoPedalSpanner(kind='sostenuto')
            >>> abjad.tweak(spanner).color = 'blue'
            >>> abjad.tweak(spanner).sostenuto_pedal_line_spanner.staff_padding = 5
            >>> abjad.attach(spanner, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \set Staff.pedalSustainStyle = #'mixed
                    c'8
                    - \tweak SostenutoPedalLineSpanner.staff-padding #5
                    - \tweak color #blue
                    \sostenutoOn
                    d'8
                    e'8
                    f'8
                    \sostenutoOff
                }

        ..  container:: example

            Una corda / tre corde pedal:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> spanner = abjad.PianoPedalSpanner(kind='corda')
            >>> abjad.attach(spanner, staff[:])
            >>> abjad.override(staff[0]).staff.una_corda_pedal.color = 'blue'
            >>> abjad.override(staff[0]).staff.una_corda_pedal_line_spanner.staff_padding = 5
            >>> abjad.override(staff[-1]).staff.una_corda_pedal.color = 'blue'
            >>> abjad.override(staff[-1]).staff.una_corda_pedal_line_spanner.staff_padding = 5
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \once \override Staff.UnaCordaPedal.color = #blue
                    \once \override Staff.UnaCordaPedalLineSpanner.staff-padding = #5
                    \set Staff.pedalSustainStyle = #'mixed
                    c'8
                    \unaCorda
                    d'8
                    e'8
                    \once \override Staff.UnaCordaPedal.color = #blue
                    \once \override Staff.UnaCordaPedalLineSpanner.staff-padding = #5
                    f'8
                    \treCorde
                }

            Note that when no visible bracket connects start-text and stop-text
            indications (as above) that the first and last leaves in spanner
            must be overriden independently. (Note also that ``abjad.tweak()``
            will apply to only the pedal-down indication and will leave the
            pedal-up indication unmodified.)

        Returns ``'sustain'``, ``'sostenuto'`` or ``'corda'``.
        """
        return self._kind

    @property
    def style(self) -> str:
        r"""
        Gets style of piano pedal spanner.

        ..  container:: example

            Mixed style:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> spanner = abjad.PianoPedalSpanner(style='mixed')
            >>> abjad.tweak(spanner).sustain_pedal_line_spanner.staff_padding = 5
            >>> abjad.attach(spanner, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \set Staff.pedalSustainStyle = #'mixed
                    c'8
                    - \tweak SustainPedalLineSpanner.staff-padding #5
                    \sustainOn
                    d'8
                    e'8
                    f'8
                    \sustainOff
                }

        ..  container:: example

            Bracket style:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> spanner = abjad.PianoPedalSpanner(style='bracket')
            >>> abjad.tweak(spanner).sustain_pedal_line_spanner.staff_padding = 5
            >>> abjad.attach(spanner, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \set Staff.pedalSustainStyle = #'bracket
                    c'8
                    - \tweak SustainPedalLineSpanner.staff-padding #5
                    \sustainOn
                    d'8
                    e'8
                    f'8
                    \sustainOff
                }

        ..  container:: example

            Text style:

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> spanner = abjad.PianoPedalSpanner(style='text')
            >>> abjad.override(staff[0]).staff.sustain_pedal_line_spanner.staff_padding = 5
            >>> abjad.override(staff[-1]).staff.sustain_pedal_line_spanner.staff_padding = 5
            >>> abjad.attach(spanner, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \once \override Staff.SustainPedalLineSpanner.staff-padding = #5
                    \set Staff.pedalSustainStyle = #'text
                    c'8
                    \sustainOn
                    d'8
                    e'8
                    \once \override Staff.SustainPedalLineSpanner.staff-padding = #5
                    f'8
                    \sustainOff
                }

            Note that when no visible bracket connects start-text and stop-text
            indications (as above) that the first and last leaves in spanner
            must be overriden independently. (Note also that ``abjad.tweak()``
            will apply to only the pedal-down indication and will leave the
            pedal-up indication unmodified.)

        Returns ``'mixed'``, ``'bracket'`` or ``'text'``.
        """
        return self._style



import collections
from .TypedCollection import TypedCollection


class OrderedDict(TypedCollection, collections.MutableMapping):
    r"""
    Ordered dictionary.

    ..  container:: example

        Initializes from list of pairs:

        >>> dictionary = abjad.OrderedDict([
        ...     ('color', 'red'),
        ...     ('directive', abjad.Markup(r'\italic Allegretto')),
        ...     ])

        >>> abjad.f(dictionary)
        abjad.OrderedDict(
            [
                ('color', 'red'),
                (
                    'directive',
                    abjad.Markup(
                        contents=[
                            abjad.MarkupCommand(
                                'italic',
                                'Allegretto'
                                ),
                            ],
                        ),
                    ),
                ]
            )

    ..  container:: example

        Initializes from built-in dictionary:

        >>> dictionary = {
        ...     'color': 'red',
        ...     'directive': abjad.Markup(r'\italic Allegretto'),
        ...     }
        >>> dictionary = abjad.OrderedDict(
        ...     dictionary
        ...     )

        >>> abjad.f(dictionary)
        abjad.OrderedDict(
            [
                ('color', 'red'),
                (
                    'directive',
                    abjad.Markup(
                        contents=[
                            abjad.MarkupCommand(
                                'italic',
                                'Allegretto'
                                ),
                            ],
                        ),
                    ),
                ]
            )

    ..  container:: example

        Initializes from other ordered dictionary:

        >>> dictionary_1 = abjad.OrderedDict([
        ...     ('color', 'red'),
        ...     ('directive', abjad.Markup(r'\italic Allegretto')),
        ...     ])
        >>> dictionary_2 = abjad.OrderedDict(
        ...     dictionary_1
        ...     )

        >>> abjad.f(dictionary_2)
        abjad.OrderedDict(
            [
                ('color', 'red'),
                (
                    'directive',
                    abjad.Markup(
                        contents=[
                            abjad.MarkupCommand(
                                'italic',
                                'Allegretto'
                                ),
                            ],
                        ),
                    ),
                ]
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        TypedCollection.__init__(
            self,
            item_class=item_class,
            )
        if isinstance(items, dict):
            items = sorted(items.items())
        elif isinstance(items, collections.Mapping):
            items = list(items.items())
        items = items or []
        the_items = []
        for item in items:
            assert len(item) == 2, repr(item)
            key = item[0]
            value = self._item_coercer(item[1])
            the_item = (key, value)
            the_items.append(the_item)
        self._collection = collections.OrderedDict(the_items)

    ### SPECIAL METHODS ###

    def __cmp__(self, argument):
        """
        Aliases OrderedDict.__cmp__().

        Returns true or false.
        """
        assert isinstance(argument, type(self))
        ordered_dictionary = argument._collection
        return self._collection.__cmp__(ordered_dictionary)

    def __contains__(self, key):
        """
        Aliases OrderedDict.__contains__().

        Returns true or false.
        """
        return key in self._collection

    def __delitem__(self, i):
        """
        Aliases OrderedDict.__delitem__().

        Returns none.
        """
        del(self._collection[i])

    def __ge__(self, argument):
        """
        Is true when typed ordered dictionary is greater than or equal
        to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__ge__(argument._collection)

    def __getitem__(self, argument):
        """
        Gets item or slice identified by ``argument``.

        Returns item or slice.
        """
        return self._collection.__getitem__(argument)

    def __gt__(self, argument):
        """
        Is true when typed ordered dictionary is greater than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__gt__(argument._collection)

    def __le__(self, argument):
        """
        Is true when typed ordered dictionary is less than or equal
        to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__le__(argument._collection)

    def __lt__(self, argument):
        """
        Is true when typed ordered dictionary is less than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__lt__(argument._collection)

    def __reversed__(self):
        """
        Aliases OrderedDict.__reversed__().

        Returns generatos.
        """
        return self._collection.__reversed__()

    def __setitem__(self, i, argument):
        """
        Changes items in ``argument`` to items and sets.

        Returns none.
        """
        new_item = self._item_coercer(argument)
        self._collection[i] = new_item

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        import abjad
        agent = abjad.StorageFormatManager(self)
        names = list(agent.signature_keyword_names)
        if 'items' in names:
            names.remove('items')
        values = [list(self._collection.items())]
        return abjad.FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=values,
            storage_format_kwargs_names=names,
            )

    ### PUBLIC METHODS ###

    def clear(self):
        """
        Clears typed ordered dictionary.

        Returns none.
        """
        self._collection.clear()

    def copy(self):
        """
        Copies typed ordered dictionary.

        Returns new typed ordered dictionary.
        """
        ordered_dictionary = self._collection.copy()
        items = list(ordered_dictionary.items())
        return type(self)(
            item_class=self.item_class,
            items=items,
            )

    def get(self, i, default=None):
        """
        Aliases OrderedDict.get().

        Returns item or raises key error.
        """
        return self._collection.get(i, default)

    def has_key(self, key):
        """
        Aliases OrderdDict.has_key().

        Returns true or false.
        """
        return key in self._collection

    def items(self):
        """
        Aliases OrderedDict.items().

        Returns generator.
        """
        return iter(self._collection.items())

    def keys(self):
        """
        Aliases OrderedDict.keys().

        Returns generator.
        """
        return iter(self._collection.keys())

    def pop(self, key, default=None):
        """
        Aliases OrderedDict.pop().

        Returns items.
        """
        return self._collection.pop(key, default)

    def popitem(self):
        """
        Aliases OrderedDict.popitem().

        Returns generator.
        """
        return self._collection.popitem()

    def setdefault(self, key, default=None):
        """
        Aliases OrderedDict.setdefault().

        Returns items.
        """
        return self._collection.setdefault(key, default)

    def sort(self, recurse=False) -> None:
        """
        Sorts ordered dictionary (in place).

        ..  container:: example

            >>> dictionary = abjad.OrderedDict()
            >>> dictionary['flavor'] = 'cherry'
            >>> dictionary['colors'] = abjad.OrderedDict()
            >>> dictionary['colors']['red'] = 3
            >>> dictionary['colors']['green'] = 2
            >>> dictionary['colors']['blue'] = 1
            >>> abjad.f(dictionary)
            abjad.OrderedDict(
                [
                    ('flavor', 'cherry'),
                    (
                        'colors',
                        abjad.OrderedDict(
                            [
                                ('red', 3),
                                ('green', 2),
                                ('blue', 1),
                                ]
                            ),
                        ),
                    ]
                )

            >>> dictionary.sort()
            >>> abjad.f(dictionary)
            abjad.OrderedDict(
                [
                    (
                        'colors',
                        abjad.OrderedDict(
                            [
                                ('red', 3),
                                ('green', 2),
                                ('blue', 1),
                                ]
                            ),
                        ),
                    ('flavor', 'cherry'),
                    ]
                )

            >>> dictionary.sort(recurse=True)
            >>> abjad.f(dictionary)
            abjad.OrderedDict(
                [
                    (
                        'colors',
                        abjad.OrderedDict(
                            [
                                ('blue', 1),
                                ('green', 2),
                                ('red', 3),
                                ]
                            ),
                        ),
                    ('flavor', 'cherry'),
                    ]
                )

        """
        items = list(self.items())
        items.sort()
        self.clear()
        for key, value in items:
            if recurse is True and isinstance(value, OrderedDict):
                value.sort(recurse=True)
            self[key] = value

    def update(self, *arguments, **keywords):
        """
        Aliases OrderedDict.update().

        Returns none.
        """
        return self._collection.update(*arguments, **keywords)

    def values(self):
        """
        Aliases OrderedDict.values().

        Returns generator.
        """
        return iter(self._collection.values())



import copy
import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from .Tag import Tag


class Wrapper(AbjadValueObject):
    r"""
    Wrapper.

    ..  container:: example

        >>> component = abjad.Note("c'4")
        >>> articulation = abjad.Articulation('accent', direction=abjad.Up)
        >>> abjad.attach(articulation, component)
        >>> wrapper = abjad.inspect(component).wrapper()

        >>> abjad.f(wrapper)
        abjad.Wrapper(
            indicator=abjad.Articulation('accent', Up),
            tag=abjad.Tag(),
            )

    ..  container:: example

        Duplicate indicator warnings take two forms.

        >>> voice_1 = abjad.Voice("c''4 d'' e'' f''", name='VoiceI')
        >>> voice_2 = abjad.Voice("c'4 d' e' f'", name='VoiceII')
        >>> abjad.attach(abjad.Clef('alto'), voice_2[0])
        >>> staff = abjad.Staff([voice_1, voice_2], is_simultaneous=True)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            <<
                \context Voice = "VoiceI"
                {
                    c''4
                    d''4
                    e''4
                    f''4
                }
                \context Voice = "VoiceII"
                {
                    \clef "alto"
                    c'4
                    d'4
                    e'4
                    f'4
                }
            >>

        First form when attempting to attach a contexted indicator to a leaf
        that already carries a contexted indicator of the same type:

        >>> abjad.attach(abjad.Clef('treble'), voice_2[0])
        Traceback (most recent call last):
            ...
        abjad...PersistentIndicatorError: Can not attach ...
        <BLANKLINE>
            abjad.Clef('treble')
        <BLANKLINE>
            ... to Note("c'4") in VoiceII because ...
        <BLANKLINE>
        abjad.Wrapper(
            context='Staff',
            indicator=abjad.Clef('alto'),
            tag=abjad.Tag(),
            )
        <BLANKLINE>
            ... is already attached to the same leaf.
        <BLANKLINE>

        Second form when attempting to attach a contexted indicator to a leaf
        governed by some other component carrying a contexted indicator of the
        same type.

        >>> abjad.attach(abjad.Clef('treble'), voice_1[0])
        Traceback (most recent call last):
            ...
        abjad...PersistentIndicatorError: Can not attach ...
        <BLANKLINE>
            abjad.Clef('treble')
        <BLANKLINE>
            ... to Note("c''4") in VoiceI because ...
        <BLANKLINE>
        abjad.Wrapper(
            context='Staff',
            indicator=abjad.Clef('alto'),
            tag=abjad.Tag(),
            )
        <BLANKLINE>
            ... is already attached to Note("c'4") in VoiceII.
        <BLANKLINE>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Internals'

    __slots__ = (
        '_alternate',
        '_annotation',
        '_component',
        '_context',
        '_deactivate',
        '_effective_context',
        '_indicator',
        '_synthetic_offset',
        '_tag',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(
        self,
        alternate: typing.Tuple[str, str] = None,
        annotation: str = None,
        component=None,
        context: str = None,
        deactivate: bool = None,
        indicator=None,
        synthetic_offset: int = None,
        tag: typing.Union[str, Tag] = None,
        ) -> None:
        from abjad.core.Component import Component
        from abjad.core.Context import Context
        from abjad.spanners.Spanner import Spanner
        from abjad.utilities.Offset import Offset
        assert not isinstance(indicator, type(self)), repr(indicator)
        if alternate is not None:
            assert isinstance(alternate, tuple) and len(alternate) == 2
        self._alternate = alternate
        if annotation is not None:
            assert isinstance(annotation, str), repr(annotation)
        self._annotation = annotation
        if component is not None:
            assert isinstance(component, (Component, Spanner))
        self._component = component
        if deactivate is not None:
            deactivate = bool(deactivate)
        if context is not None:
            assert isinstance(context, str), repr(context)
        self._context = context
        if deactivate is not None:
            deactivate = bool(deactivate)
        self._deactivate = deactivate
        self._effective_context = None
        self._indicator = indicator
        if synthetic_offset is not None:
            synthetic_offset = Offset(synthetic_offset)
        self._synthetic_offset = synthetic_offset
        if tag is not None:
            assert isinstance(tag, (str, Tag))
        tag = Tag(tag)
        self._tag: Tag = tag

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments) -> 'Wrapper':
        r"""
        Copies indicator wrapper.

        ..  container:: example

            Preserves annotation flag:

            >>> old_staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> abjad.annotate(old_staff[0], 'bow_direction', abjad.Down)
            >>> abjad.f(old_staff)
            \new Staff {
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = old_staff[0]
            >>> abjad.inspect(leaf).annotation('bow_direction')
            Down

            >>> new_staff = abjad.mutate(old_staff).copy()
            >>> abjad.f(new_staff)
            \new Staff {
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = new_staff[0]
            >>> abjad.inspect(leaf).annotation('bow_direction')
            Down

        ..  container:: example

            Preserves tag:

            >>> old_staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> clef = abjad.Clef('alto')
            >>> abjad.attach(clef, old_staff[0], tag='RED:M1')
            >>> abjad.f(old_staff)
            \new Staff {
                \clef "alto" %! RED:M1
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = old_staff[0]
            >>> wrapper = abjad.inspect(leaf).wrapper()
            >>> abjad.f(wrapper)
            abjad.Wrapper(
                context='Staff',
                indicator=abjad.Clef('alto'),
                tag=abjad.Tag('RED:M1'),
                )

            >>> new_staff = abjad.mutate(old_staff).copy()
            >>> abjad.f(new_staff)
            \new Staff {
                \clef "alto" %! RED:M1
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = new_staff[0]
            >>> wrapper = abjad.inspect(leaf).wrapper()
            >>> abjad.f(wrapper)
            abjad.Wrapper(
                context='Staff',
                indicator=abjad.Clef('alto'),
                tag=abjad.Tag('RED:M1'),
                )

        ..  container:: example

            Preserves deactivate flag:

            >>> old_staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> abjad.attach(
            ...     abjad.Clef('alto'),
            ...     old_staff[0],
            ...     deactivate=True,
            ...     tag='RED:M1',
            ...     )
            >>> abjad.f(old_staff)
            \new Staff {
                %@% \clef "alto" %! RED:M1
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = old_staff[0]
            >>> wrapper = abjad.inspect(leaf).wrapper()
            >>> abjad.f(wrapper)
            abjad.Wrapper(
                context='Staff',
                deactivate=True,
                indicator=abjad.Clef('alto'),
                tag=abjad.Tag('RED:M1'),
                )

            >>> new_staff = abjad.mutate(old_staff).copy()
            >>> abjad.f(new_staff)
            \new Staff {
                %@% \clef "alto" %! RED:M1
                c'4
                d'4
                e'4
                f'4
            }

            >>> leaf = new_staff[0]
            >>> wrapper = abjad.inspect(leaf).wrapper()
            >>> abjad.f(wrapper)
            abjad.Wrapper(
                context='Staff',
                deactivate=True,
                indicator=abjad.Clef('alto'),
                tag=abjad.Tag('RED:M1'),
                )

        Copies all properties except component.
        
        Copy operations must supply component after wrapper copy.

        Returns new indicator wrapper.
        """
        new = type(self)(
            alternate=self.alternate,
            annotation=self.annotation,
            component=None,
            context=self.context,
            deactivate=self.deactivate,
            indicator=copy.copy(self.indicator),
            synthetic_offset=self.synthetic_offset,
            tag=self.tag,
            )
        return new

    ### PRIVATE METHODS ###

    def _bind_correct_effective_context(self, correct_effective_context):
        import abjad
        self._unbind_effective_context()
        if correct_effective_context is not None:
            correct_effective_context._dependent_wrappers.append(self)
        self._effective_context = correct_effective_context
        self._update_effective_context()
        if isinstance(self.indicator, abjad.MetronomeMark):
            correct_effective_context._update_later(offsets_in_seconds=True)

    def _bind_to_component(self, component):
        import abjad
        self._warn_duplicate_indicator(component)
        self._unbind_component()
        self._component = component
        self._update_effective_context()
        if isinstance(self.indicator, abjad.MetronomeMark):
            self._component._update_later(offsets_in_seconds=True)
        component._wrappers.append(self)

    def _detach(self):
        self._unbind_component()
        self._unbind_effective_context()
        return self

    def _find_correct_effective_context(self):
        import abjad
        if self.context is None:
            return None
        context = getattr(abjad, self.context, self.context)
        if isinstance(context, type):
            for component in abjad.inspect(self.component).parentage():
                if not isinstance(component, abjad.Context):
                    continue
                if isinstance(component, context):
                    return component
        elif isinstance(context, str):
            for component in abjad.inspect(self.component).parentage():
                if not isinstance(component, abjad.Context):
                    continue
                if (component.name == context or
                    component.lilypond_type == context):
                    return component
        else:
            raise TypeError('must be context or string: {context!r}.')

    def _get_effective_context(self):
        if self.component is not None:
            self.component._update_now(indicators=True)
        return self._effective_context

    def _get_format_pieces(self):
        import abjad
        result = []
        if self.annotation:
            return result
        if hasattr(self.indicator, '_get_lilypond_format_bundle'):
            bundle = self.indicator._get_lilypond_format_bundle(self.component)
            bundle.tag_format_contributions(
                self.tag,
                deactivate=self.deactivate,
                )
            return bundle
        try:
            context = self._get_effective_context()
            lilypond_format = self.indicator._get_lilypond_format(
                context=context,
                )
        except TypeError:
            lilypond_format = self.indicator._get_lilypond_format()
        if isinstance(lilypond_format, str):
            lilypond_format = [lilypond_format]
        assert isinstance(lilypond_format, (tuple, list))
        lilypond_format = abjad.LilyPondFormatManager.tag(
            lilypond_format,
            self.tag,
            deactivate=self.deactivate,
            )
        result.extend(lilypond_format)
        if self._get_effective_context() is not None:
            return result
        if isinstance(self.indicator, abjad.TimeSignature):
            if isinstance(self.component, abjad.Measure):
                return result
        result = [rf'%%% {_} %%%' for _ in result]
        return result

    def _get_format_specification(self):
        import abjad
        keywords = [
            'alternate',
            'annotation',
            'context',
            'deactivate',
            'indicator',
            'synthetic_offset',
            'tag',
            ]
        return abjad.FormatSpecification(
            client=self,
            storage_format_args_values=None,
            storage_format_kwargs_names=keywords,
            )

    def _is_formattable_for_component(self, component):
        import abjad
        if self.annotation:
            return False
        if isinstance(self.component, abjad.Measure):
            if self.component is component:
                if not isinstance(self.indicator, abjad.TimeSignature):
                    return True
                elif component.always_format_time_signature:
                    return True
                else:
                    previous_measure = self.component._get_previous_measure()
                    if previous_measure is not None:
                        previous_effective_time_signature = \
                            previous_measure.time_signature
                    else:
                        previous_effective_time_signature = None
                    if not self.indicator == previous_effective_time_signature:
                        return True
        elif getattr(self.indicator, '_format_slot', None) == 'right':
            if self.component is component:
                return True
        elif self.component is component:
            return True
        return False

    def _unbind_component(self):
        component = self.component
        if component is not None:
            if hasattr(component, '_wrappers'):
                if self in component._wrappers:
                    component._wrappers.remove(self)
        self._component = None

    def _unbind_effective_context(self):
        effective_context = self._effective_context
        if effective_context is not None:
            try:
                effective_context._dependent_wrappers.remove(self)
            except ValueError:
                pass
        self._effective_context = None

    def _update_effective_context(self):
        current_effective_context = self._effective_context
        correct_effective_context = self._find_correct_effective_context()
        if current_effective_context is not correct_effective_context:
            self._bind_correct_effective_context(correct_effective_context)

    def _warn_duplicate_indicator(self, component):
        import abjad
        if isinstance(component, abjad.Spanner):
            return
        if isinstance(self.indicator, abjad.LilyPondLiteral):
            return
        if self.deactivate is True:
            return
        prototype = type(self.indicator)
        wrapper = abjad.inspect(component).effective_wrapper(prototype)
        if (wrapper is not None and
            wrapper.context is not None and
            wrapper.deactivate is not True and
            wrapper.indicator != self.indicator):
            if wrapper.start_offset == self.start_offset:
                parentage = abjad.inspect(component).parentage()
                context = parentage.get_first(abjad.Context)
                message = f'\n\nCan not attach ...\n\n{self.indicator}\n\n...'
                message += f' to {repr(component)}'
                message += f' in {context.name} because ...'
                message += f'\n\n{format(wrapper)}\n\n'
                message += '... is already attached'
                if component is wrapper.component:
                    message += ' to the same leaf.'
                else:
                    inspection = abjad.inspect(wrapper.component)
                    parentage = inspection.parentage()
                    context = parentage.get_first(abjad.Context)
                    message += f' to {repr(wrapper.component)}'
                    message += f' in {context.name}.'
                message += '\n'
                message = message.format(
                    self.indicator,
                    repr(component),
                    format(wrapper),
                    repr(wrapper.component),
                    )
                raise abjad.PersistentIndicatorError(message)

    ### PUBLIC PROPERTIES ###

    @property
    def alternate(self) -> typing.Optional[typing.Tuple[str, str]]:
        """
        Gets alternate tagging information.

        Set only by ``MetronomeMarkSpanner.attach(..., alternate=None)``
        keyword.

        Returns (color, tag) pair, or none.
        """
        return self._alternate

    @property
    def annotation(self) -> typing.Optional[str]:
        """
        Gets indicator wrapper annotation.

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> articulation = abjad.Articulation('accent', direction=abjad.Up)
            >>> abjad.attach(articulation, note)
            >>> wrapper = abjad.inspect(note).wrapper()
            >>> wrapper.annotation is None
            True

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> articulation = abjad.Articulation('accent', direction=abjad.Up)
            >>> abjad.annotate(note, 'foo', articulation)
            >>> abjad.inspect(note).annotation('foo')
            Articulation('accent', Up)

        """
        return self._annotation

    @property
    def component(self):
        """
        Gets start component of indicator wrapper.

        Returns component.
        """
        from abjad.spanners.Spanner import Spanner
        if isinstance(self._component, Spanner):
            if self._component:
                return self._component[0]
            else:
                return None
        return self._component

    @property
    def context(self) -> typing.Optional[str]:
        """
        Gets context of indicator wrapper.
        """
        return self._context

    @property
    def deactivate(self) -> typing.Optional[bool]:
        """
        Is true when wrapper deactivates tag.
        """
        assert self._deactivate in (True, False, None)
        return self._deactivate

    @deactivate.setter
    def deactivate(self, argument):
        assert argument in (True, False, None)
        self._deactivate: typing.Optional[bool] = argument

    @property
    def indicator(self) -> typing.Any:
        """
        Gets indicator of indicator wrapper.
        """
        return self._indicator

    @property
    def start_offset(self):
        """
        Gets start offset of indicator wrapper.

        This is either the wrapper's synthetic offset or the start offset of
        the wrapper's component.

        Returns offset.
        """
        from abjad.top.inspect import inspect
        if self._synthetic_offset is not None:
            return self._synthetic_offset
        return inspect(self._component).timespan().start_offset

    @property
    def synthetic_offset(self):
        """
        Gets synthetic offset of indicator wrapper.

        Returns offset or none.
        """
        return self._synthetic_offset

    @property
    def tag(self) -> Tag:
        """
        Gets and sets tag.
        """
        assert isinstance(self._tag, Tag), repr(self._tag)
        return self._tag

    @tag.setter
    def tag(self, argument):
        if not isinstance(argument, (str, Tag)):
            raise Exception(f'string or tag: {argument!r}.')
        tag = Tag(argument)
        self._tag = tag



import typing
from abjad.utilities.String import String
from .LilyPondNameManager import LilyPondNameManager


class LilyPondSettingNameManager(LilyPondNameManager):
    """
    LilyPond setting name manager.

    ..  container:: example

        >>> note = abjad.Note("c'4")
        >>> abjad.setting(note)
        LilyPondSettingNameManager()

    """

    ### SPECIAL METHODS ###

    def __getattr__(self, name:str) -> typing.Any:
        r"""
        Gets arbitrary object keyed to ``name``.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.setting(staff).instrument_name = abjad.Markup('Vn. I')
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    instrumentName = \markup { "Vn. I" }
                }
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }

        ..  container:: example

            Returns arbitrary object keyed to ``name``:

            >>> abjad.setting(staff).instrument_name
            Markup(contents=['Vn. I'])

        """
        from abjad.ly import contexts
        camel_name = String(name).to_upper_camel_case()
        if name.startswith('_'):
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                message = '{type_name!r} object has no attribute: {name!r}.'
                raise AttributeError(message)
        elif camel_name in contexts:
            try:
                return vars(self)['_' + name]
            except KeyError:
                context = LilyPondNameManager()
                vars(self)['_' + name] = context
                return context
        else:
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                message = '{type_name!r} object has no attribute: {name!r}.'
                raise AttributeError(message)

    ### PRIVATE METHODS ###

    def _get_attribute_tuples(self):
        result = []
        for name, value in vars(self).items():
            if type(value) is LilyPondNameManager:
                prefixed_context_name = name
                lilypond_type = prefixed_context_name.strip('_')
                context_proxy = value
                attribute_pairs = context_proxy._get_attribute_pairs()
                for attribute_name, attribute_value in attribute_pairs:
                    triple = (lilypond_type, attribute_name, attribute_value)
                    result.append(triple)
            else:
                attribute_name, attribute_value = name, value
                result.append((attribute_name, attribute_value))
        return result



from .TimeRelation import TimeRelation


class OffsetTimespanTimeRelation(TimeRelation):
    """
    Offfset vs. timespan time relation.

    ..  container:: example

        >>> offset = abjad.Offset(5)
        >>> timespan = abjad.Timespan(0, 10)
        >>> time_relation = abjad.timespans.offset_happens_during_timespan(
        ...     offset=offset,
        ...     timespan=timespan,
        ...     hold=True,
        ...     )

        ::

            >>> abjad.f(time_relation)
            abjad.timespans.OffsetTimespanTimeRelation(
                inequality=abjad.timespans.CompoundInequality(
                    [
                        abjad.TimespanInequality('timespan.start <= offset'),
                        abjad.TimespanInequality('offset < timespan.stop'),
                        ],
                    logical_operator='and',
                    ),
                timespan=abjad.Timespan(
                    start_offset=abjad.Offset(0, 1),
                    stop_offset=abjad.Offset(10, 1),
                    ),
                offset=abjad.Offset(5, 1),
                )

    Offset / timespan time relations are immutable.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Time relations'

    __slots__ = (
        '_offset',
        '_timespan',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, inequality=None, timespan=None, offset=None):
        TimeRelation.__init__(self, inequality=inequality)
        self._timespan = timespan
        self._offset = offset

    ### SPECIAL METHODS ###

    # TODO: hoist to TimeRelation
    def __call__(self, timespan=None, offset=None):
        """
        Evaluates time relation:

            >>> offset = abjad.Offset(5)
            >>> timespan = abjad.Timespan(0, 10)
            >>> time_relation = abjad.timespans.offset_happens_during_timespan(
            ...     offset=offset,
            ...     timespan=timespan,
            ...     hold=True,
            ...     )
            >>> time_relation()
            True

        Raises value error is either ``offset`` or ``timespan`` is none.

        Otherwise returns boolean.
        """
        import abjad
        timespan = timespan or self.timespan
        offset = offset or self.offset
        if timespan is None or offset is None:
            message = 'time relation is not fully loaded.'
            raise ValueError(message)
        if not isinstance(timespan, abjad.Timespan):
            timespan = abjad.Timespan()._get_timespan(timespan)
        offset = abjad.Offset(offset)
        truth_value = self.inequality.evaluate_offset_inequality(
            timespan.start_offset, timespan.stop_offset, offset)
        return truth_value

    def __eq__(self, argument):
        """
        Is true when ``argument`` equals time relation.

        ..  container:: example

            >>> offset = abjad.Offset(5)
            >>> time_relation_1 = abjad.timespans.offset_happens_during_timespan()
            >>> time_relation_2 = abjad.timespans.offset_happens_during_timespan(
            ...     offset=offset,
            ...     )

            >>> time_relation_1 == time_relation_1
            True
            >>> time_relation_1 == time_relation_2
            False
            >>> time_relation_2 == time_relation_2
            True

        Returns true or false.
        """
        return super().__eq__(argument)

    def __format__(self, format_specification=''):
        """
        Formats time relation.

        ::

            >>> offset = abjad.Offset(5)
            >>> timespan = abjad.Timespan(0, 10)
            >>> time_relation = abjad.timespans.offset_happens_during_timespan(
            ...     offset=offset,
            ...     timespan=timespan,
            ...     hold=True,
            ...     )
            >>> abjad.f(time_relation)
            abjad.timespans.OffsetTimespanTimeRelation(
                inequality=abjad.timespans.CompoundInequality(
                    [
                        abjad.TimespanInequality('timespan.start <= offset'),
                        abjad.TimespanInequality('offset < timespan.stop'),
                        ],
                    logical_operator='and',
                    ),
                timespan=abjad.Timespan(
                    start_offset=abjad.Offset(0, 1),
                    stop_offset=abjad.Offset(10, 1),
                    ),
                offset=abjad.Offset(5, 1),
                )

        Returns string.
        """
        return super().__format__(format_specification=format_specification)

    def __hash__(self):
        """
        Hashes time relation.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    ### PUBLIC PROPERTIES ###

    @property
    def is_fully_loaded(self):
        """
        Is true when ``timespan`` and ``offset`` are both not none.

        ..  container:: example

            >>> offset = abjad.Offset(5)
            >>> timespan = abjad.Timespan(0, 10)
            >>> time_relation = abjad.timespans.offset_happens_during_timespan(
            ...     offset=offset,
            ...     timespan=timespan,
            ...     hold=True,
            ...     )
            >>> time_relation.is_fully_loaded
            True

        Returns true or false.
        """
        return self.timespan is not None and self.offset is not None

    @property
    def is_fully_unloaded(self):
        """
        Is true when ``timespan`` and ``offset`` are both none.

        ..  container:: example

            >>> offset = abjad.Offset(5)
            >>> timespan = abjad.Timespan(0, 10)
            >>> time_relation = abjad.timespans.offset_happens_during_timespan(
            ...     offset=offset,
            ...     timespan=timespan,
            ...     hold=True,
            ...     )
            >>> time_relation.is_fully_unloaded
            False

        Returns true or false.
        """
        return self.timespan is None and self.offset is None

    @property
    def offset(self):
        """
        Time relation offset:

        ::

            >>> offset = abjad.Offset(5)
            >>> timespan = abjad.Timespan(0, 10)
            >>> time_relation = abjad.timespans.offset_happens_during_timespan(
            ...     offset=offset,
            ...     timespan=timespan,
            ...     hold=True,
            ...     )
            >>> time_relation.offset
            Offset(5, 1)

        Returns offset or none.
        """
        return self._offset

    @property
    def timespan(self):
        """
        Time relation timepsan:

        ::

            >>> offset = abjad.Offset(5)
            >>> timespan = abjad.Timespan(0, 10)
            >>> time_relation = abjad.timespans.offset_happens_during_timespan(
            ...     offset=offset,
            ...     timespan=timespan,
            ...     hold=True,
            ...     )
            >>> time_relation.timespan
            Timespan(start_offset=Offset(0, 1), stop_offset=Offset(10, 1))

        Returns timespan or none.
        """
        return self._timespan



import typing
from abjad.core.Chord import Chord
from abjad.core.Note import Note
from abjad.indicators.BendAfter import BendAfter
from abjad.lilypondnames.LilyPondGrobOverride import LilyPondGrobOverride
from abjad.scheme import SchemeSymbol
from abjad.top.inspect import inspect
from abjad.utilities.OrderedDict import OrderedDict
from .Spanner import Spanner


class Glissando(Spanner):
    r"""
    Glissando.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> glissando = abjad.Glissando()
        >>> abjad.attach(glissando, staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'8
                \glissando
                d'8
                \glissando
                e'8
                \glissando
                f'8
            }

    ..  container:: example

        Glissando avoids bend-after indicators:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> bend_after = abjad.BendAfter()
        >>> abjad.attach(bend_after, staff[1])
        >>> glissando = abjad.Glissando()
        >>> abjad.attach(glissando, staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'8
                \glissando
                d'8
                - \bendAfter #'-4
                e'8
                \glissando
                f'8
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_allow_repeats',
        '_allow_ties',
        '_parenthesize_repeats',
        '_right_broken',
        '_stems',
        '_style',
        )

    _start_command = r'\glissando'

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        allow_repeats: bool = None,
        allow_ties: bool = None,
        parenthesize_repeats: bool = None,
        right_broken: bool = None,
        stems: bool = None,
        style: str = None,
        ) -> None:
        Spanner.__init__(self)
        if allow_repeats is not None:
            allow_repeats = bool(allow_repeats)
        self._allow_repeats = allow_repeats
        if allow_ties is not None:
            allow_ties = bool(allow_ties)
        self._allow_ties = allow_ties
        if parenthesize_repeats is not None:
            parenthesize_repeats = bool(parenthesize_repeats)
        self._parenthesize_repeats = parenthesize_repeats
        if right_broken is not None:
            right_broken = bool(right_broken)
        self._right_broken = right_broken
        if stems is not None:
            stems = bool(stems)
        self._stems = stems
        if style is not None:
            assert isinstance(style, str), repr(style)
        self._style = style

    ### PRIVATE METHODS ###

    def _copy_keywords(self, new):
        Spanner._copy_keywords(self, new)
        new._allow_repeats = self.allow_repeats
        new._allow_ties = self.allow_ties
        new._parenthesize_repeats = self.parenthesize_repeats

    def _get_lilypond_format_bundle(self, leaf):
        bundle = self._get_basic_lilypond_format_bundle(leaf)
        prototype = (Chord, Note)
        should_attach_glissando = False
        if leaf is not self[0]:
            if self.parenthesize_repeats:
                if not self._previous_leaf_changes_current_pitch(leaf):
                    self._parenthesize_leaf(leaf)
        tag = False
        if inspect(leaf).has_indicator(BendAfter):
            pass
        elif leaf is self[-1]:
            if self._right_broken:
                should_attach_glissando = True
                tag = True
        elif not isinstance(leaf, prototype):
            pass
        elif self.allow_repeats and self.allow_ties:
            should_attach_glissando = True
        elif self.allow_repeats and not self.allow_ties:
            should_attach_glissando = self._is_last_in_tie_chain(leaf)
        elif not self.allow_repeats and self.allow_ties:
            if self._next_leaf_changes_current_pitch(leaf):
                should_attach_glissando = True
        elif (not self.allow_repeats and not self.allow_ties):
            if self._next_leaf_changes_current_pitch(leaf):
                if self._is_last_in_tie_chain(leaf):
                    should_attach_glissando = True
        if self.stems:
            if leaf is self[1]:
                strings = [
                    r'\override Accidental.stencil = ##f',
                    r'\override NoteColumn.glissando-skip = ##t',
                    r'\hide NoteHead',
                    r'\override NoteHead.no-ledgers = ##t',
                    ]
                bundle.grob_overrides.extend(strings)
            if leaf is self[-1]:
                strings = [
                    r'\revert Accidental.stencil',
                    r'\revert NoteColumn.glissando-skip',
                    r'\undo \hide NoteHead',
                    r'\revert NoteHead.no-ledgers',
                    ]
                if self._right_broken:
                    strings_ = self._tag_hide(strings)
                    bundle.grob_reverts.extend(strings_)
                    strings_ = self._tag_show(strings)
                    bundle.after.commands.extend(strings_)
                else:
                    bundle.grob_reverts.extend(strings)
        if self.style:
            if leaf is self[0]:
                override = LilyPondGrobOverride(
                    grob_name='Glissando',
                    property_path='style',
                    value=SchemeSymbol(self.style),
                    )
                string = override.tweak_string()
                bundle.after.spanner_starts.append(string)
        if should_attach_glissando:
            strings = self._tweaked_start_command_strings()
            if tag:
                strings = self._tag_show(strings)
            bundle.after.spanner_starts.extend(strings)
        return bundle

    @staticmethod
    def _is_last_in_tie_chain(leaf):
        logical_tie = inspect(leaf).logical_tie()
        return leaf is logical_tie[-1]

    @staticmethod
    def _next_leaf_changes_current_pitch(leaf):
        next_leaf = inspect(leaf).leaf(n=1)
        if (isinstance(leaf, Note) and
            isinstance(next_leaf, Note) and
            leaf.written_pitch == next_leaf.written_pitch):
            return False
        elif (isinstance(leaf, Chord) and
            isinstance(next_leaf, Chord) and
            leaf.written_pitches == next_leaf.written_pitches):
            return False
        return True

    @staticmethod
    def _parenthesize_leaf(leaf):
        if isinstance(leaf, Note):
            leaf.note_head.is_parenthesized = True
        elif isinstance(leaf, Chord):
            for note_head in leaf.note_heads:
                note_head.is_parenthesized = True

    @staticmethod
    def _previous_leaf_changes_current_pitch(leaf):
        previous_leaf = inspect(leaf).leaf(n=-1)
        if (isinstance(leaf, Note) and
            isinstance(previous_leaf, Note) and
            leaf.written_pitch == previous_leaf.written_pitch):
            return False
        elif (isinstance(leaf, Chord) and
            isinstance(previous_leaf, Chord) and
            leaf.written_pitches == previous_leaf.written_pitches):
            return False
        return True

    ### PUBLIC PROPERTIES ###

    @property
    def allow_repeats(self) -> typing.Optional[bool]:
        r"""
        Is true when glissando should allow repeated pitches.

        ..  container:: example

            Does not allow repeated pitches:

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     allow_repeats=False,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    a8
                    \glissando
                    b8
                    ~
                    b8
                    \glissando
                    c'8
                    c'8
                    \glissando
                    d'8
                    ~
                    d'8
                }

            This is default behavior.

        ..  container:: example

            Allows repeated pitches (but not ties):

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     allow_repeats=True,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    \glissando
                    a8
                    \glissando
                    b8
                    ~
                    b8
                    \glissando
                    c'8
                    \glissando
                    c'8
                    \glissando
                    d'8
                    ~
                    d'8
                }

        ..  container:: example

            Allows both repeated pitches and ties:

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     allow_repeats=True,
            ...     allow_ties=True,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    \glissando
                    a8
                    \glissando
                    b8
                    ~
                    \glissando
                    b8
                    \glissando
                    c'8
                    \glissando
                    c'8
                    \glissando
                    d'8
                    ~
                    \glissando
                    d'8
                }

        Ties are excluded when repeated pitches are not allowed because all
        ties comprise repeated pitches.
        """
        return self._allow_repeats

    @property
    def allow_ties(self) -> typing.Optional[bool]:
        r"""
        Is true when glissando should allow ties.

        ..  container:: example

            Does not allow repeated pitches (including ties):

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     allow_repeats=False,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    a8
                    \glissando
                    b8
                    ~
                    b8
                    \glissando
                    c'8
                    c'8
                    \glissando
                    d'8
                    ~
                    d'8
                }

            This is default behavior.

        ..  container:: example

            Allows repeated pitches (but not ties):

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     allow_repeats=True,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    \glissando
                    a8
                    \glissando
                    b8
                    ~
                    b8
                    \glissando
                    c'8
                    \glissando
                    c'8
                    \glissando
                    d'8
                    ~
                    d'8
                }

        ..  container:: example

            Allows both repeated pitches and ties:

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     allow_repeats=True,
            ...     allow_ties=True,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    \glissando
                    a8
                    \glissando
                    b8
                    ~
                    \glissando
                    b8
                    \glissando
                    c'8
                    \glissando
                    c'8
                    \glissando
                    d'8
                    ~
                    \glissando
                    d'8
                }

        Ties are excluded when repeated pitches are not allowed because all
        ties comprise repeated pitches.
        """
        return self._allow_ties

    @property
    def cross_segment_examples(self):
        r"""
        Cross-segment examples.

        ..  container:: example

            Cross-segment example #1 (one-to-one):

            >>> segment_1 = abjad.Voice("c'4 d' e' f'", name='MainVoice')
            >>> glissando = abjad.Glissando(right_broken=True)
            >>> abjad.attach(
            ...     glissando,
            ...     segment_1[-1:],
            ...     )
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    e'4
                    f'4
                %@% \glissando                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("g'4 a' b' r", name='MainVoice')
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    g'4
                    a'4
                    b'4
                    r4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP
            
            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        e'4
                        f'4
                        \glissando                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        g'4
                        a'4
                        b'4
                        r4
                    }
                }

        ..  container:: example

            Cross-segment example #2 (one-to-many):

            >>> segment_1 = abjad.Voice("c'4 d' e' f'", name='MainVoice')
            >>> glissando = abjad.Glissando(right_broken=True)
            >>> abjad.attach(glissando, segment_1[-1:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    d'4
                    e'4
                    f'4
                %@% \glissando                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("g'4 a' b' r", name='MainVoice')
            >>> abjad.attach(abjad.Glissando(), segment_2[:3])
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    g'4
                    \glissando
                    a'4
                    \glissando
                    b'4
                    r4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP
            
            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        d'4
                        e'4
                        f'4
                        \glissando                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        g'4
                        \glissando
                        a'4
                        \glissando
                        b'4
                        r4
                    }
                }

        ..  container:: example

            Cross-segment example #3 (many-to-one):

            >>> segment_1 = abjad.Voice("c'4 d' e' f'", name='MainVoice')
            >>> glissando = abjad.Glissando(right_broken=True)
            >>> abjad.attach(glissando, segment_1[:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    \glissando
                    d'4
                    \glissando
                    e'4
                    \glissando
                    f'4
                %@% \glissando                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("g'4 a' b' r", name='MainVoice')
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    g'4
                    a'4
                    b'4
                    r4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP
            
            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        \glissando
                        d'4
                        \glissando
                        e'4
                        \glissando
                        f'4
                        \glissando                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        g'4
                        a'4
                        b'4
                        r4
                    }
                }

        ..  container:: example

            Cross-segment example #4 (many-to-many):

            >>> segment_1 = abjad.Voice("c'4 d' e' f'", name='MainVoice')
            >>> glissando = abjad.Glissando(right_broken=True)
            >>> abjad.attach(glissando, segment_1[:])
            >>> abjad.show(segment_1, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_1, strict=50)
                \context Voice = "MainVoice"
                {
                    c'4
                    \glissando
                    d'4
                    \glissando
                    e'4
                    \glissando
                    f'4
                %@% \glissando                                    %! SHOW_TO_JOIN_BROKEN_SPANNERS
                }

            >>> segment_2 = abjad.Voice("g'4 a' b' r", name='MainVoice')
            >>> abjad.attach(abjad.Glissando(), segment_2[:3])
            >>> abjad.show(segment_2, strict=50) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(segment_2, strict=50)
                \context Voice = "MainVoice"
                {
                    g'4
                    \glissando
                    a'4
                    \glissando
                    b'4
                    r4
                }

            >>> container = abjad.Container([segment_1, segment_2])
            >>> text = format(container, 'lilypond')
            >>> text = abjad.LilyPondFormatManager.left_shift_tags(text, 50)
            >>> job = abjad.Job.join_broken_spanners(text)
            >>> text = job()
            >>> lines = text.split('\n')
            >>> lilypond_file = abjad.LilyPondFile.new(lines)
            >>> abjad.show(lilypond_file) # doctest: +SKIP
            
            ..  docs::

                >>> print(text)
                {
                    \context Voice = "MainVoice"
                    {
                        c'4
                        \glissando
                        d'4
                        \glissando
                        e'4
                        \glissando
                        f'4
                        \glissando                                %! SHOW_TO_JOIN_BROKEN_SPANNERS %@%
                    }
                    \context Voice = "MainVoice"
                    {
                        g'4
                        \glissando
                        a'4
                        \glissando
                        b'4
                        r4
                    }
                }

        """
        pass

    @property
    def parenthesize_repeats(self) -> typing.Optional[bool]:
        r"""
        Is true when glissando should parenthesize repeated pitches.

        ..  container:: example

            Does not parenthesize repeated pitches:

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando()
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    a8
                    \glissando
                    b8
                    ~
                    b8
                    \glissando
                    c'8
                    c'8
                    \glissando
                    d'8
                    ~
                    d'8
                }

            This is default behavior.

        ..  container:: example

            Spans and parenthesizes repeated pitches:

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     allow_repeats=True,
            ...     parenthesize_repeats=True,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    \glissando
                    \parenthesize
                    a8
                    \glissando
                    b8
                    ~
                    \parenthesize
                    b8
                    \glissando
                    c'8
                    \glissando
                    \parenthesize
                    c'8
                    \glissando
                    d'8
                    ~
                    \parenthesize
                    d'8
                }

        ..  container:: example

            Parenthesizes (but does not span) repeated pitches:

            >>> staff = abjad.Staff("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8")
            >>> glissando = abjad.Glissando(
            ...     parenthesize_repeats=True,
            ...     )
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    a8
                    \parenthesize
                    a8
                    \glissando
                    b8
                    ~
                    \parenthesize
                    b8
                    \glissando
                    c'8
                    \parenthesize
                    c'8
                    \glissando
                    d'8
                    ~
                    \parenthesize
                    d'8
                }

        """
        return self._parenthesize_repeats

    @property
    def right_broken(self) -> typing.Optional[bool]:
        """
        Is true when spanner is right-broken.
        """
        return self._right_broken

    @property
    def stems(self) -> typing.Optional[bool]:
        r"""
        Is true when glissando formats stems-only timing marks non nonedge
        leaves.

        ..  container:: example

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> glissando = abjad.Glissando(stems=True)
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    \glissando
                    \hide NoteHead
                    \override Accidental.stencil = ##f
                    \override NoteColumn.glissando-skip = ##t
                    \override NoteHead.no-ledgers = ##t
                    d'8
                    \glissando
                    e'8
                    \glissando
                    \revert Accidental.stencil
                    \revert NoteColumn.glissando-skip
                    \revert NoteHead.no-ledgers
                    \undo \hide NoteHead
                    f'8
                }

        """
        return self._stems

    @property
    def style(self) -> typing.Optional[str]:
        r"""
        Gets style.

        ..  container:: example

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> glissando = abjad.Glissando(style='trill')
            >>> abjad.attach(glissando, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    c'8
                    - \tweak style #'trill
                    \glissando
                    d'8
                    \glissando
                    e'8
                    \glissando
                    f'8
                }

        """
        return self._style



import typing
from abjad.pitch.PitchSegment import PitchSegment
from .Spanner import Spanner


class OctavationSpanner(Spanner):
    r"""
    Octavation spanner.

    ..  container:: example

        Spans four notes:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> spanner = abjad.OctavationSpanner(start=1)
        >>> abjad.attach(spanner, staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \ottava #1
                c'4
                d'4
                e'4
                f'4
                \ottava #0
            }

    ..  container:: example

        Spans one note:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> spanner = abjad.OctavationSpanner(start=1)
        >>> abjad.attach(spanner, staff[:1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \ottava #1
                c'4
                \ottava #0
                d'4
                e'4
                f'4
            }

        One-note octavation changes are allowed.

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_start',
        '_stop',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        start: int = 1,
        stop: int = 0,
        ) -> None:
        Spanner.__init__(self)
        assert isinstance(start, int)
        self._start = start
        assert isinstance(stop, int)
        self._stop = stop

    ### PRIVATE PROPERTIES ###

    @property
    def _start_command(self):
        return rf'\ottava #{self.start}'

    @property
    def _stop_command(self):
        return rf'\ottava #{self.stop}'

    ### PRIVATE METHODS ###

    def _copy_keywords(self, new):
        new._start = self.start
        new._stop = self.stop

    def _get_lilypond_format_bundle(self, leaf):
        bundle = self._get_basic_lilypond_format_bundle(leaf)
        if leaf is self[0]:
            strings = self._tweaked_start_command_strings()
            bundle.before.commands.extend(strings)
        if leaf is self[-1]:
            string = self._stop_command_string()
            bundle.after.commands.append(string)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def start(self) -> typing.Optional[int]:
        """
        Gets octavation start.

        ..  container:: example

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
            >>> spanner = abjad.OctavationSpanner(start=1)
            >>> abjad.attach(spanner, staff[:])
            >>> abjad.show(staff) # doctest: +SKIP

            >>> spanner.start
            1

        """
        return self._start

    @property
    def stop(self) -> typing.Optional[int]:
        """
        Gets octavation stop.

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> spanner = abjad.OctavationSpanner(start=2, stop=1)
        >>> abjad.attach(spanner, staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        >>> spanner.stop
        1

        """
        return self._stop



import math


def divisors(n):
    """
    Gets positive divisors of ``n`` in increasing order.

    ..  container:: example

        >>> abjad.mathtools.divisors(84)
        [1, 2, 3, 4, 6, 7, 12, 14, 21, 28, 42, 84]

        >>> for x in range(10, 20):
        ...     print(x, abjad.mathtools.divisors(x))
        ...
        10 [1, 2, 5, 10]
        11 [1, 11]
        12 [1, 2, 3, 4, 6, 12]
        13 [1, 13]
        14 [1, 2, 7, 14]
        15 [1, 3, 5, 15]
        16 [1, 2, 4, 8, 16]
        17 [1, 17]
        18 [1, 2, 3, 6, 9, 18]
        19 [1, 19]

    ..  container:: example

        Allows nonpositive ``n``:

        >>> abjad.mathtools.divisors(-27)
        [1, 3, 9, 27]

    Raises not implemented error on ``0``.

    Returns list of positive integers.
    """
    if not isinstance(n, int):
        message = 'must be integer: {!r}.'
        message = message.format(n)
        raise TypeError(message)
    if n == 0:
        message = 'all numbers divide zero evenly.'
        raise NotImplementedError(message)
    n = abs(n)
    divisors = [1]
    for i in range(2, int(math.sqrt(n)) + 1):
        if n % i == 0:
            divisors.append(i)
    codivisors = [n // i for i in reversed(divisors)]
    if divisors[-1] == codivisors[0]:
        divisors.pop()
    divisors.extend(codivisors)
    divisors.sort()
    return divisors
